from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from fnmatch import fnmatch
from typing import Any, Generic, Iterable, TypeVar

from ._address import Address
from ._addressable import Addressable

T_Addressable = TypeVar("T_Addressable", bound=Addressable)


class AddressSpace(ABC, Generic[T_Addressable]):
    """
    This is a base class to represent a space of #Addressable objects. Its methods must implement the navigation
    within that space per the elements in an #Address.

    See also: #resolve_address()
    """

    @abstractmethod
    def get_root(self) -> T_Addressable:
        """
        Returns the root node of the addressable space.
        """

        raise NotImplementedError(self)

    @abstractmethod
    def get_parent(self, entity: T_Addressable) -> T_Addressable | None:
        """
        Returns the parent of an addressable object.
        """

        raise NotImplementedError(self)

    @abstractmethod
    def get_children(self, entity: T_Addressable) -> Iterable[T_Addressable]:
        """
        Returns the children of an addressable object.
        """

        raise NotImplementedError(self)


@dataclass
class AddressResolutionStep(Generic[T_Addressable]):
    """
    A step in the address resolution. Each step represents the object and address at which the resolution
    continues to the next element of the address. When the last element of the address is reached, the
    #matches are filled.
    """

    #: The addressable object which served as the starting point for this step in the address resolution.
    entity: T_Addressable

    #: The remaining address that is resolved in this step in the context of #entity.
    query: Address

    #: A list of matches from the #entity that matched the #query. This is only filled if either the #query
    #: is the root address or a single element in the address remains; denoting the leaf and ultimate target
    #: of the query.
    #:
    #: See also: #is_leaf()
    matches: list[T_Addressable]

    #: A reference to the previous step in the resolution.
    previous_step: AddressResolutionStep[T_Addressable] | None = field(repr=False)

    #: A list of zero or more next steps in the resolution. This is only filled if the resolution step is not
    #: a leaf.
    #:
    #: See also: is_leaf()
    next_steps: list[AddressResolutionStep[T_Addressable]]

    def is_leaf(self) -> bool:
        """
        Returns `True` if the #query address indicates that this resolution step is a leaf step. The leaf
        step contains matching objects in the addressable space in #matches and contains no #next_steps.

        A leaf is indicated by the #query address either

        * being the root address (`:`), or
        * not being absolute and containing exactly one element (e.g. `a`, `.`, `**` and vice versa)

        >>> AddressResolutionStep(None, Address(":a"), [], None, []).is_leaf()
        False
        >>> AddressResolutionStep(None, Address("a"), [], None, []).is_leaf()
        True
        >>> AddressResolutionStep(None, Address("a:b"), [], None, []).is_leaf()
        False
        """

        return self.query.is_root() or (not self.query.is_absolute() and len(self.query) == 1)

    def is_concrete(self) -> bool:
        """
        Returns `True` if the resolution step is concrete and should thus have exactly one element in #matches (if
        the step is a leaf) or one element in #next_steps (if the step is not a leaf). A concrete step is one where

        * the #query address is either absolute (where the next steps or matches leads to the root of the addressable
          space), or
        * the first element is concrete (see #Address.Element.is_concrete()), or
        * the previous previous step was a recursive wildcard (`**`)

        >>> AddressResolutionStep(None, Address(":a?"), [], None, []).is_concrete()
        True
        >>> AddressResolutionStep(None, Address("a*"), [], None, []).is_concrete()
        False
        >>> AddressResolutionStep(None, Address("a:b?"), [], None, []).is_concrete()
        True
        >>> previous_step = AddressResolutionStep(None, Address("**"), [], None, [])
        >>> AddressResolutionStep(None, Address("b"), [], previous_step, []).is_concrete()
        False
        """

        if (
            self.previous_step
            and len(self.previous_step.query) > 0
            and self.previous_step.query[0].is_recursive_wildcard()
        ):
            return False

        return self.query.is_absolute() or bool(self.query and self.query[0].is_concrete())


@dataclass
class AddressResolutionResult(Generic[T_Addressable]):
    """
    A wrapper for the resolution of address resolution using #resolve_address().

    In addition to containing a reference to the initial resolution step in #root, it provides
    utility methods for working with the resulting tree.
    """

    space: AddressSpace[T_Addressable]
    root: AddressResolutionStep[T_Addressable]

    def all_steps(self) -> Iterable[AddressResolutionStep[T_Addressable]]:
        """
        Returns a generator that iterates over all resolution steps.
        """

        def generator(step: AddressResolutionStep[T_Addressable]) -> Iterable[AddressResolutionStep[T_Addressable]]:
            yield step
            for next_step in step.next_steps:
                yield from generator(next_step)

        return generator(self.root)

    def matches(self) -> Iterable[T_Addressable]:
        """
        Returns all #AddressResolutionStep.matches from all steps in the tree.
        """

        for step in self.all_steps():
            yield from step.matches


def resolve_address(
    space: AddressSpace[T_Addressable],
    entity: T_Addressable,
    query: Address,
) -> AddressResolutionResult[T_Addressable]:
    """
    Resolves an address in any address *space*, starting from any addressable object in that space *entity* to
    find zero or more related entities in the same space by following the specified *query* address.

    In essence, this function follows the address elements in the *query* starting from *entity* and returns the
    matching entities.

    >>> from ._address_resolver_test import Node, NodeAddressSpace
    >>> root = Node(None, "root")
    >>> a = Node(root, "a")
    >>> aa = Node(a, "a")
    >>> c = Node(root, "c")
    >>> ca = Node(c, "a")
    >>> space = NodeAddressSpace(root)
    >>>
    >>> list(resolve_address(space, aa, Address(":a")).matches())
    [Node(address=Address(':a'))]
    >>> list(resolve_address(space, c, Address(":a:a")).matches())
    [Node(address=Address(':a:a'))]
    >>> list(resolve_address(space, c, Address("..:a")).matches())
    [Node(address=Address(':a'))]
    >>> list(resolve_address(space, root, Address("*:a")).matches())
    [Node(address=Address(':a:a')), Node(address=Address(':c:a'))]
    >>> list(resolve_address(space, root, Address("d?")).matches())
    []

    When an address cannot be resolved, you get an #AddressResolutionError.

    >>> resolve_address(space, root, Address("d"))
    Traceback (most recent call last):
    kraken.core.address._address_resolver.AddressResolutionError: Could not resolve address 'd' in context \
':'. The failure occurred at address ':' trying to resolve the remainder 'd'. The address ':d' does not exist.

    Recursive wildcards expect to find at least one match on the next element, unless that element is fallible:

    >>> list(resolve_address(space, root, Address("**:c")).matches())
    [Node(address=Address(':c'))]
    >>> resolve_address(space, root, Address("**:d"))
    Traceback (most recent call last):
    kraken.core.address._address_resolver.AddressResolutionError: Could not resolve address '**:d' in context \
':'. The failure occurred at address ':' trying to resolve the remainder '**:d'. The address ':**:d' does not exist.
    >>> list(resolve_address(space, root, Address("**:d?")).matches())
    []
    """

    root_entity = entity
    root_query = query

    if query.is_empty():
        raise ValueError("An empty address query cannot be resolved")

    def recurse_tree(entity: T_Addressable, include_root: bool = True) -> Iterable[T_Addressable]:
        if include_root:
            yield entity
        for child in space.get_children(entity):
            yield from recurse_tree(child)

    def has_children(
        space: AddressSpace[T_Addressable],
        addr: T_Addressable,
    ) -> bool:
        return any(True for _ in space.get_children(addr))

    def resolve_step(
        previous_step: AddressResolutionStep[T_Addressable] | None,
        entity: T_Addressable,
        query: Address,
        restrict_to_containers: bool = False,
    ) -> AddressResolutionStep[T_Addressable]:
        assert not query.is_empty()

        current_step = AddressResolutionStep[T_Addressable](
            entity=entity,
            query=query,
            matches=[],
            previous_step=previous_step,
            next_steps=[],
        )

        if query.is_absolute():
            element = None
            next_entities = [space.get_root()]
            if len(query.elements) == 0:
                remainder = None
            else:
                remainder = Address.create(False, query.is_container(), query.elements)
        else:
            element = query.elements[0]
            if len(query.elements) > 1:
                remainder = Address.create(query.is_absolute(), query.is_container(), query.elements[1:])
            else:
                remainder = Address("")
            if element.is_current():
                next_entities = [entity]
            elif element.is_parent():
                next_entities = list(filter(None, (space.get_parent(entity),)))
            elif element.is_recursive_wildcard():
                # TODO(NiklasRosenstein): We might want to keep the generator here for perf/mem.
                # NOTE: When we have no more elements to resolve following the recursive wildcard, we don't
                #   want the result here to include the current entity (e.g. `:a:**` should not include `:a`).
                next_entities = list(recurse_tree(entity, include_root=not remainder.is_empty()))
            else:
                next_entities = [
                    x  # linebreak for formatting
                    for x in space.get_children(entity)
                    if fnmatch(x.address.name, element.value)
                ]

        # Check if the resolution for this step failed. Only when the step was resolving a concrete element
        # of the address can the resolution fail. If the previous element was a recursive wildcard (`**`) or if
        # the current element is fallible (`a?`) then we permit no results.
        if current_step.is_concrete() and not next_entities:
            raise AddressResolutionError(space, root_entity, root_query, entity, query)

        if not remainder:
            # There are no more elements to continue branching the address resolution for.
            if restrict_to_containers:
                current_step.matches += [node for node in next_entities if has_children(space, node)]
            else:
                current_step.matches += next_entities
        else:
            # There are elements to continue the address resolution for.
            for next_entity in next_entities:
                next_step = resolve_step(current_step, next_entity, remainder, restrict_to_containers)
                current_step.next_steps.append(next_step)

            # If the current element is a recursive wildcard, and the next element is not fallible, we require
            # that at least one matching element was found in the next steps of the resolution.
            if not remainder[-1].fallible and element and element.is_recursive_wildcard():
                if not any(s.next_steps or s.matches for s in current_step.next_steps):
                    raise AddressResolutionError(space, root_entity, root_query, entity, query)

        return current_step

    restrict_to_containers = query.is_container()
    return AddressResolutionResult(space=space, root=resolve_step(None, entity, query, restrict_to_containers))


@dataclass
class AddressResolutionError(Exception):
    """
    This exception is raised by #resolve_address() if a concrete resolution step failed.
    """

    space: AddressSpace[Any]
    entity: Addressable
    query: Address
    failed_at: Addressable
    remainder: Address

    def is_recursive_wildcard_failure(self) -> bool:
        """
        Returns `True` if this error is caused by a failure to resolve at least one member of a recursive wildcard
        element in the address. This is the case when the #remainder's first element is a recursive wildcard. Recursive
        wildcards themselves cannot fail to resolve, but only their immediate following element.

        For example, when `Address(":**:d")` is resolved and no `d` exists, then this method returns `True` for
        the exception that is raised by #resolve_address().
        """

        return len(self.remainder) > 0 and self.remainder[0].is_recursive_wildcard()

    def get_nonexistent_address(self) -> Address:
        """
        Calculates the absolute address that does not exist to illustrate the point of failure in the address
        resolution. For failures to resolve the next element of a recursive wildcard, this will return the
        address including the wildcard, e.g. `:**:d` would indicate that in the entire address space there exists
        not a single addressable object where the #Address.name is `d`.
        """

        shift = 1 if self.is_recursive_wildcard_failure() else 0
        clipped = Address.create(
            self.remainder.is_absolute(),
            self.remainder.is_container(),
            self.remainder.elements[: shift + int(not self.remainder.is_absolute())],
        )
        return self.failed_at.address.concat(clipped)

    def __str__(self) -> str:
        return (
            f"Could not resolve address '{self.query}' in context '{self.entity.address}'. "
            f"The failure occurred at address '{self.failed_at.address}' trying to resolve the remainder "
            f"'{self.remainder}'. The address '{self.get_nonexistent_address()}' does not exist."
        )



from __future__ import annotations

import contextlib
import logging
import tempfile
from pathlib import Path
from typing import Iterator

import pytest

from kraken.core.system.context import Context
from kraken.core.system.project import Project

__all__ = [
    "kraken_ctx",
    "kraken_project",
]

logger = logging.getLogger(__name__)


@pytest.fixture(name="kraken_ctx")
def _kraken_ctx_fixture() -> Iterator[Context]:
    with kraken_ctx() as ctx:
        yield ctx


@contextlib.contextmanager
def kraken_ctx() -> Iterator[Context]:
    with tempfile.TemporaryDirectory() as tmpdir:
        context = Context(Path(tmpdir))
        with context.as_current():
            yield context


@pytest.fixture(name="kraken_project")
def _kraken_project_fixture(kraken_ctx: Context) -> Iterator[Project]:
    with kraken_project(kraken_ctx) as project:
        yield project


@contextlib.contextmanager
def kraken_project(kraken_ctx: Context) -> Iterator[Project]:
    with tempfile.TemporaryDirectory() as tmpdir:
        kraken_ctx.root_project = Project("test", Path(tmpdir), None, kraken_ctx)
        with kraken_ctx.root_project.as_current():
            yield kraken_ctx.root_project



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from kraken.core.address import Address

    from .project import Project


class ProjectNotFoundError(Exception):
    def __init__(self, address: Address) -> None:
        self.address = address

    def __str__(self) -> str:
        return f"project not found: {self.address}"


class ProjectLoaderError(Exception):
    def __init__(self, project: Project, message: str) -> None:
        self.project = project
        self.message = message

    def __str__(self) -> str:
        return f"[{self.project.address}] {self.message}"


class BuildError(Exception):
    def __init__(self, failed_tasks: Iterable[str]) -> None:
        self.failed_tasks = set(failed_tasks)

    def __repr__(self) -> str:
        if len(self.failed_tasks) == 1:
            return f'task "{next(iter(self.failed_tasks))}" failed'
        else:
            return "tasks " + ", ".join(f'"{task}"' for task in sorted(self.failed_tasks)) + " failed"



from __future__ import annotations

import logging
import re
from pathlib import Path
from typing import Iterable

import dill  # type: ignore[import]
from kraken.common import pluralize

from kraken.core import Context, TaskGraph

logger = logging.getLogger(__name__)
state_file_regex = r"^state-.*\.dill$"
state_file_template = "state-{name}.dill"


def load_build_state(state_dirs: Iterable[Path]) -> tuple[Context, TaskGraph] | tuple[None, None]:
    # Find all state files that match the state filename format.
    state_files = []
    for state_dir in state_dirs:
        if state_dir.is_dir():
            for path in state_dir.iterdir():
                if re.match(state_file_regex, path.name):
                    state_files.append(path)
    if not state_files:
        return None, None

    logger.info(
        "Resuming from %d build %s (%s)",
        len(state_files),
        pluralize("state", state_files),
        ", ".join(file.name for file in state_files),
    )

    context: Context | None = None
    graph: TaskGraph | None = None

    for state_file in sorted(state_files):
        with state_file.open("rb") as fp:
            new_graph: TaskGraph = dill.load(fp)
        if context is None or graph is None:
            # We want to retrieve the entire, original build graph.
            context, graph = new_graph.context, new_graph.root
        # Take the results from the final graph (only this graph contains results).
        graph.results_from(new_graph)

    assert context is not None and graph is not None
    return context, graph


def save_build_state(state_dir: Path, name: str, graph: TaskGraph) -> None:
    state_file = state_dir / state_file_template.format(name=name)
    state_dir.mkdir(parents=True, exist_ok=True)
    with state_file.open("wb") as fp:
        dill.dump(graph, fp)
    for file in state_dir.iterdir():
        if file != state_file:
            file.unlink()
    logger.info('Saving build state to "%s"', state_file)



"""
Provides generic base classes that are relative independent of Kraken itself.
"""

from .currentable import Currentable, CurrentProvider
from .metadata import MetadataContainer

__all__ = [
    "Currentable",
    "CurrentProvider",
    "MetadataContainer",
]



from __future__ import annotations

from typing import Collection

from termcolor import colored as _colored

from kraken.core import Graph, Task, TaskGraph, TaskStatus, TaskStatusType
from kraken.core.system.executor.default import DefaultPrintingExecutorObserver

COLORS_BY_STATUS = {
    TaskStatusType.PENDING: "magenta",
    TaskStatusType.FAILED: "red",
    TaskStatusType.INTERRUPTED: "red",
    TaskStatusType.SKIPPED: "yellow",
    TaskStatusType.WARNING: "yellow",
    TaskStatusType.SUCCEEDED: "green",
    TaskStatusType.STARTED: "magenta",
    TaskStatusType.UP_TO_DATE: "green",
}


def status_to_text(status: TaskStatus, colored: bool = True) -> str:
    if colored:
        message = _colored(status.type.name, COLORS_BY_STATUS.get(status.type))
    else:
        message = status.type.name
    if status.message:
        message += f" ({status.message})"
    return message


class ColoredDefaultPrintingExecutorObserver(DefaultPrintingExecutorObserver):
    def __init__(
        self,
        exclude_tasks: Collection[Task] | None = None,
        exclude_task_subgraphs: Collection[Task] | None = None,
    ) -> None:
        super().__init__(
            status_to_text=status_to_text,
            format_header=lambda s: _colored(s, "cyan", attrs=["bold", "underline"]),
            format_duration=lambda s: _colored(s, "cyan"),
        )
        self.exclude_tasks = exclude_tasks
        self.exclude_task_subgraphs = exclude_task_subgraphs

    def _mark_tasks_as_skipped(self, graph: TaskGraph, tasks: Collection[Task], recursive: bool = False) -> None:
        for task in tasks:
            if not graph.get_status(task):
                status = TaskStatus.skipped("excluded")
                graph.set_status(task, status)
                self.after_execute_task(task, status)
            if recursive:
                self._mark_tasks_as_skipped(graph, graph.get_predecessors(task, ignore_groups=False), recursive=True)

    # GraphExecutorObserver

    def before_execute_graph(self, graph: Graph) -> None:
        super().before_execute_graph(graph)
        assert isinstance(
            graph, TaskGraph
        ), f"{type(self).__name__} expected a TaskGraph instance, got {type(graph).__name__}"
        self._mark_tasks_as_skipped(graph, self.exclude_tasks or [])
        self._mark_tasks_as_skipped(graph, self.exclude_task_subgraphs or [], True)



"""
This module provides the :class:`Address` class and related classes and methods to work with addresses and
addressable objects.
"""

from ._address import Address
from ._address_resolver import AddressResolutionError, AddressResolutionResult, AddressSpace, resolve_address
from ._addressable import Addressable

__all__ = [
    "Address",
    "AddressResolutionResult",
    "Addressable",
    "AddressSpace",
    "AddressResolutionError",
    "resolve_address",
]



from __future__ import annotations

from pathlib import Path
from typing import List, Optional, Tuple, Union

from kraken.common.supplier import OfSupplier, VoidSupplier
from pytest import mark, raises
from typing_extensions import Literal

from kraken.core.system.property import Property, PropertyContainer


def test__Property_value_adapter_order_is_semantically_revelant() -> None:
    """Tests that a `str | Path` and `Path | str` property behave differently."""

    prop1: Property[str | Path] = Property(PropertyContainer(), "prop1", Union[str, Path])
    prop1.set("foo/bar")
    assert prop1.get() == "foo/bar"

    prop2: Property[Path | str] = Property(PropertyContainer(), "prop2", Union[Path, str])
    prop2.set("foo/bar")
    assert prop2.get() == Path("foo/bar")

    prop2.setmap(lambda s: Path(str(s).upper()))
    assert prop2.get() == Path("FOO/BAR")


def test__Property_default() -> None:
    """Tests that property defaults work as expected."""

    a_value = ["abc"]

    class MyObj(PropertyContainer):
        a: Property[list[str]] = Property.default(a_value)
        b: Property[int] = Property.default_factory(lambda: 42)
        c: Property[str]

    obj = MyObj()
    assert obj.a.get() == a_value
    assert obj.a.get() is not a_value  # Copied
    assert obj.b.get() == 42
    assert obj.c.is_empty()


def test__Property_default_factory_with_subclass() -> None:
    """Tests that property default factory works with a subclass (a known previous semantic failure case)."""

    class MyObj(PropertyContainer):
        b: Property[dict[str, str]] = Property.default_factory(dict)

    class SubObj(MyObj):
        pass

    obj = MyObj()
    assert obj.b.get() == {}

    subobj = SubObj()
    assert subobj.b.get() == {}


def test__Property__provides() -> None:
    assert Property[str](PropertyContainer(), "foo", str).provides(str)
    assert not Property[Path](PropertyContainer(), "foo", Path).provides(str)
    assert Property[Union[str, Path]](PropertyContainer(), "foo", Union[str, Path]).provides(str)
    assert Property[Union[str, Path]](PropertyContainer(), "foo", Union[str, Path]).provides(Path)
    assert not Property[Union[str, Path]](PropertyContainer(), "foo", Union[str, Path]).provides(int)
    assert not Property[Union[str, Path]](PropertyContainer(), "foo", Union[str, Path]).provides(type(None))
    assert Property[Optional[str]](PropertyContainer(), "foo", Optional[str]).provides(str)
    assert Property[Optional[str]](PropertyContainer(), "foo", Optional[str]).provides(type(None))


def test__Property__get_of_type__scalar() -> None:
    p1 = Property[str](PropertyContainer(), "foo", str)
    p1.set("bar")
    assert p1.get_of_type(str) == ["bar"]


def test__Property__get_of_type__scalar_no_match() -> None:
    p1 = Property[str](PropertyContainer(), "foo", str)
    p1.set("bar")
    assert p1.get_of_type(int) == []


def test__Property__get_of_type__sequence() -> None:
    p1 = Property[List[str]](PropertyContainer(), "foo", List[str])
    p1.set(["hello", "world"])
    assert p1.get_of_type(str) == ["hello", "world"]


def test__Property__get_of_type__sequence_no_match() -> None:
    p1 = Property[List[str]](PropertyContainer(), "foo", List[str])
    p1.set(["hello", "world"])
    assert p1.get_of_type(int) == []


def test__Property__get_of_type__sequence_partial_match() -> None:
    p1 = Property[List[Union[str, int]]](PropertyContainer(), "foo", List[Union[str, int]])
    p1.set(["hello", 42, "world"])
    assert p1.get_of_type(int) == [42]
    assert p1.get_of_type(str) == ["hello", "world"]
    assert p1.get_of_type(object) == ["hello", 42, "world"]
    assert p1.get_of_type(type(None)) == []
    assert p1.get_of_type(list) == [["hello", 42, "world"]]


def test__Property__deferred_exception_propagates() -> None:
    p1 = Property[str](PropertyContainer(), "foo", str)
    p2 = Property[str](PropertyContainer(), "bar", str, deferred=True)

    with raises(Property.Empty):
        p1.get()
    with raises(Property.Deferred):
        p2.get()

    p1.set(p2)
    with raises(Property.Deferred):
        p1.get()

    p2.set("hello")
    assert p1.get() == "hello"


def test__Property__deferred_registers_as_empty() -> None:
    p1 = Property[str](PropertyContainer(), "foo", str, deferred=True)
    assert p1.is_empty()
    assert not p1.is_filled()


def test__Property__property_pointing_to_a_deferred_property_does_not_register_as_empty() -> None:
    """
    A property that is populated with the value of another Property that is deferred should not register as empty,
    instead it should propagate the deferred exception.
    """

    p1 = Property[str](PropertyContainer(), "foo", str, deferred=True)
    p2 = Property[str](PropertyContainer(), "bar", str)

    p2.set(p1)
    with raises(Property.Deferred):
        p2.is_empty()

    p1.set("hello")
    assert not p2.is_empty()
    assert p2.is_filled()


def test__Property__accepts_tuple() -> None:
    p1 = Property[Tuple[str, int]](PropertyContainer(), "foo", Tuple[str, int])
    p1.set(("hello", 42))
    assert p1.get() == ("hello", 42)
    with raises(TypeError):
        p1.set(["hello", 42])  # type: ignore[arg-type]


@mark.skip(reason="we don't currently do this level of validaton")
def test__Property__does_not_accept_tuple_with_wrong_type() -> None:
    p1 = Property[Tuple[str, int]](PropertyContainer(), "foo", Tuple[str, int])
    with raises(TypeError):
        p1.set(("hello", "world"))  # type: ignore[arg-type]


def test__Property__does_accept_literal_type_hint() -> None:
    prop = Property[Literal["foo", "bar"]](PropertyContainer(), "foo", Literal["foo", "bar"])
    prop.set("foo")
    prop.set("bar")

    # Cannot set int when only string literals are accepted.
    with raises(TypeError) as excinfo:
        prop.set(42)  # type: ignore[arg-type]
    assert str(excinfo.value).endswith("expected str, got int")

    # Cannot set a string literal that is not in the list of accepted literals.
    # This is because we don't currently validate the values.
    # TODO(@NiklasRosenstein): Validate Literal values.
    prop.set("bazinga")  # type: ignore[arg-type]


def test__PropertyContainer__output_property_is_deferred_by_default() -> None:
    class MyObj(PropertyContainer):
        a: Property[str]
        b: Property[str] = Property.output()

    obj = MyObj()
    with raises(Property.Empty):
        obj.a.get()
    with raises(Property.Deferred):
        obj.b.get()

    obj.a.set("foo")
    assert obj.a.get() == "foo"
    obj.b.set("bar")
    assert obj.b.get() == "bar"


def test__PropertyContainer__descriptor_get_and_set() -> None:
    class MyObj(PropertyContainer):
        a: Property[str]
        b: Property[str] = Property.output()
        c: Property[str | None]

    assert isinstance(MyObj.a, Property)
    assert isinstance(MyObj().a, Property)
    assert MyObj.a is not MyObj().a
    assert MyObj.a.name == "a"
    assert MyObj().a.name == "a"

    obj = MyObj()
    obj.a = "foo"
    obj.b = "bar"
    assert obj.a.get() == "foo"
    assert obj.b.get() == "bar"
    with raises(Property.Empty):
        MyObj.a.get()

    # Setting a non-optional property to None sets it to a void supplier.
    print("@@>>", obj.a)
    print("@@>>", obj.a.accepted_types)
    obj.a = None
    print("@@>>", obj.a._value)
    assert isinstance(obj.a._value, VoidSupplier)

    # Setting an optional property to None just sets it to None.
    obj.c = "foo"
    assert obj.c.get() == "foo"
    obj.c = None
    assert isinstance(obj.c._value, OfSupplier)

    obj = MyObj()
    with raises(Property.Empty):
        obj.a.get()



import pytest

from kraken.core.system.graph import TaskGraph
from kraken.core.system.project import Project
from kraken.core.system.task import TaskStatus, VoidTask


def test__TaskGraph__populate(kraken_project: Project) -> None:
    task_a = kraken_project.task("a", VoidTask, group="g")
    task_b = kraken_project.task("b", VoidTask, group="g")
    group = kraken_project.group("g")

    graph = TaskGraph(kraken_project.context, False)
    graph.populate([group])

    assert set(graph.tasks()) == {group, task_a, task_b}
    assert set(graph.tasks(goals=True)) == {group}


def test__TaskGraph__trim(kraken_project: Project) -> None:
    task_a = kraken_project.task("a", VoidTask, group="g")
    task_b = kraken_project.task("b", VoidTask, group="g")
    group = kraken_project.group("g")

    graph = TaskGraph(kraken_project.context).trim([group])

    assert set(graph.tasks()) == {group, task_a, task_b}
    assert set(graph.tasks(goals=True)) == {group}

    # Trimming should have the same result as a fresh populate.
    fresh_graph = TaskGraph(kraken_project.context, populate=False)
    fresh_graph.populate([group])
    assert fresh_graph._digraph.nodes == graph._digraph.nodes
    assert fresh_graph._digraph.edges == graph._digraph.edges


def test__TaskGraph__trim_with_nested_groups(kraken_project: Project) -> None:
    task_a = kraken_project.task("a", VoidTask, group="g1")
    task_b = kraken_project.task("b", VoidTask, group="g2")
    group_1 = kraken_project.group("g1")
    group_2 = kraken_project.group("g2")
    group_1.add(group_2)

    graph = TaskGraph(kraken_project.context).trim([group_1, group_2])

    assert set(graph.tasks()) == {group_2, group_1, task_a, task_b}
    assert set(graph.tasks(goals=True)) == {group_1}

    # Trimming should have the same result as a fresh populate.
    fresh_graph = TaskGraph(kraken_project.context, populate=False)
    fresh_graph.populate([group_1])
    assert fresh_graph._digraph.nodes == graph._digraph.nodes
    assert fresh_graph._digraph.edges == graph._digraph.edges


def test__TaskGraph__ready_on_successful_completion(kraken_project: Project) -> None:
    """Tests if :meth:`TaskGraph.ready` and :meth:`TaskGraph.is_complete` work as expected.

    ```
    A -----> B -----> C
    ```
    """

    task_a = kraken_project.task("a", VoidTask)
    task_b = kraken_project.task("b", VoidTask)
    task_c = kraken_project.task("c", VoidTask)

    task_c.depends_on(task_b)
    task_b.depends_on(task_a)

    graph = TaskGraph(kraken_project.context).trim([task_c])

    assert set(graph.tasks()) == {task_c, task_b, task_a}
    assert list(graph.execution_order()) == [task_a, task_b, task_c]

    # Complete tasks one by one.
    remainder = [task_a, task_b, task_c]
    while remainder:
        assert list(graph.execution_order()) == remainder
        task = remainder.pop(0)
        assert not graph.is_complete()
        assert list(graph.ready()) == [task]
        graph.set_status(task, TaskStatus.succeeded())

    assert graph.is_complete()
    assert list(graph.ready()) == []


def test__TaskGraph__ready_on_failure(kraken_project: Project) -> None:
    """This test tests if the task delivers the correct ready tasks if a task in the graph fails.

    ```
    A        B
    |        |
    v        v
    C -----> D
    ```

    If A succeeds but B fails, C would still be executable, but D stays dormant.
    """

    task_a = kraken_project.task("a", VoidTask)
    task_b = kraken_project.task("b", VoidTask)
    task_c = kraken_project.task("c", VoidTask)
    task_d = kraken_project.task("d", VoidTask)

    task_d.depends_on(task_b)
    task_d.depends_on(task_c)
    task_c.depends_on(task_a)

    graph = TaskGraph(kraken_project.context).trim([task_d])
    assert set(graph.tasks()) == {task_d, task_b, task_c, task_a}
    assert list(graph.execution_order()) in ([task_a, task_b, task_c, task_d], [task_b, task_a, task_c, task_d])
    assert set(graph.ready()) == {task_a, task_b}

    # After B fails we can still run A.
    graph.set_status(task_b, TaskStatus.failed())
    assert list(graph.ready()) == [task_a]

    # After A is successful we can still run C.
    graph.set_status(task_a, TaskStatus.succeeded())
    assert list(graph.ready()) == [task_c]

    # D cannot continue because B has failed.
    graph.set_status(task_c, TaskStatus.succeeded())
    assert list(graph.ready()) == []
    assert not graph.is_complete()


def test__TaskGraph__ready_2(kraken_project: Project) -> None:
    """
    ```
    pythonBuild -----> pythonPublish -----> publish (group)
     \\-----> build (group)
    ```
    """

    pythonBuild = kraken_project.task("pythonBuild", VoidTask, group="build")
    pythonPublish = kraken_project.task("pythonPublish", VoidTask, group="publish")
    pythonPublish.depends_on(pythonBuild)

    publish = kraken_project.group("publish")
    graph = TaskGraph(kraken_project.context).trim([publish])
    assert list(graph.ready()) == [pythonBuild]


def test__TaskGraph__correct_execution_order_on_optional_intermediate_task(kraken_project: Project) -> None:
    """Test that the TaskGraph produces the correct order in a scenario where two tasks that need to be
    executed have another that does not need to be executed in between.

    ```
    pythonInstall --------------------------> pytest
    \\---> jtd.python ===> gen ---> (X) build - - ->/
    ```

    The expected order here is that the `pytest` task can only run after `gen`. It is important to note that
    the `pytest` task only depends optionally on the `build` task, otherwise of course running `pytest` would
    require that `build` is executed as well.

    Legend:

    * `- ->`: optional dependency
    * `--->`: strict dependency
    * `===>`: member of group
    """

    python_install = kraken_project.task("pythonInstall", VoidTask)
    jtd_python = kraken_project.task("jtd.python", VoidTask, group="gen")
    gen = kraken_project.group("gen")
    build = kraken_project.group("build")
    pytest = kraken_project.task("pytest", VoidTask)

    pytest.depends_on(python_install)
    pytest.depends_on(build, mode="order-only")
    build.depends_on(gen)
    jtd_python.depends_on(python_install)

    graph = TaskGraph(kraken_project.context)

    assert list(graph.trim([pytest, gen]).execution_order()) == [python_install, jtd_python, gen, pytest]

    assert list(graph.trim([pytest, build]).execution_order()) == [python_install, jtd_python, gen, build, pytest]


@pytest.mark.parametrize("inverse", [False, True])
def test__TaskGraph__test_inverse_group_relationship(kraken_project: Project, inverse: bool) -> None:
    """Tests that the dependency propagation between members of task groups works as expected.

    Consider two groups A and B. When B depends on A, the task graph automatically expands that dependency
    to the members of B such that each depend on the members of A. This fact is stored on the edge using
    the "implicit" marker (i.e. the relationship between the tasks was not explicit using direct task
    relationships).

    :param inverse: Whether the relationship between the groups should be expressed using an inverse relationship.
        `A -> B` should yield the same result as `B <- A`.
    """

    from kraken.core.system.graph import _Edge

    a = kraken_project.group("a")
    b = kraken_project.group("b")
    ta1 = kraken_project.task("ta1", VoidTask, group=a)
    ta2 = kraken_project.task("ta2", VoidTask, group=a)
    tb1 = kraken_project.task("tb1", VoidTask, group=b)

    if inverse:
        a.required_by(b)
    else:
        b.depends_on(a)

    graph = TaskGraph(kraken_project.context)
    assert graph.get_edge(ta1, a) == _Edge(True, False)
    assert graph.get_edge(ta2, a) == _Edge(True, False)
    assert graph.get_edge(tb1, b) == _Edge(True, False)
    assert graph.get_edge(a, b) == _Edge(True, False)

    # Implicit propagated edges.
    assert graph.get_edge(a, tb1) == _Edge(True, True)

    assert list(graph.trim([b]).execution_order()) == [ta1, ta2, a, tb1, b]


def test__TaskGraph__allow_subsequent_group_execution_on_non_strict_failed_tasks(kraken_project: Project) -> None:
    """Tests that the TaskGraph correctly allows the graph execution to continue if non-strict dependencies fail, but
    correctly returns that the graph is not complete.

    Consider two groups A and B, where B depends on A but not strictly. A should be executed before B, but if any
    tasks in A fail the TaskGraph should still allow B to be executed afterwards as the dependency is not strict. The
    TaskGraph however should not consider the graph to be `complete` after B finishes executing as there still is a
    failed task.
    """
    a = kraken_project.group("a")
    b = kraken_project.group("b")

    ta1 = kraken_project.task("ta1", VoidTask, group=a)
    ta2 = kraken_project.task("ta2", VoidTask, group=a)
    tb1 = kraken_project.task("tb1", VoidTask, group=b)

    b.depends_on(a, mode="order-only")
    graph = TaskGraph(kraken_project.context)
    assert list(graph.ready()) == [ta1, ta2]

    graph.set_status(ta1, TaskStatus.succeeded())
    graph.set_status(ta2, TaskStatus.failed())

    assert list(graph.ready()) == [tb1]

    graph.set_status(tb1, TaskStatus.succeeded())
    assert not graph.is_complete()


def test__TaskGraph__allow_subsequent_task_execution_on_non_strict_failed_tasks(kraken_project: Project) -> None:
    """Tests that the TaskGraph correctly allows the graph execution to continue if non-strict dependencies fail, but
    correctly returns that the graph is not complete.

    Consider two tasks within the same group, A1 and A2. A1 should be executed before A2, but if A1 fails the TaskGraph
    should still allow A2 to be executed afterwards as the dependency is not strict. The TaskGraph however should not
    consider the graph to be `complete` after A2 finishes executing as there still is a failed task.
    """
    a = kraken_project.group("a")

    ta1 = kraken_project.task("ta1", VoidTask, group=a)
    ta2 = kraken_project.task("ta2", VoidTask, group=a)

    ta2.depends_on(ta1, mode="order-only")
    graph = TaskGraph(kraken_project.context)
    assert list(graph.ready()) == [ta1]

    graph.set_status(ta1, TaskStatus.failed())

    assert list(graph.ready()) == [ta2]

    graph.set_status(ta2, TaskStatus.succeeded())
    assert not graph.is_complete()


def test__TaskGraph__wait_for_group_finish_before_removing_non_strict_dependencies(kraken_project: Project) -> None:
    """Tests that the TaskGraph correctly waits for all tasks within a group to succeed or fail, before removing
    any non-strict dependencies to allow subsequent tasks to execute.
    """
    a = kraken_project.group("a")
    b = kraken_project.group("b")

    ta1 = kraken_project.task("ta1", VoidTask, group=a)
    ta2 = kraken_project.task("ta2", VoidTask, group=a)
    tb1 = kraken_project.task("tb1", VoidTask, group=b)

    b.depends_on(a, mode="order-only")
    graph = TaskGraph(kraken_project.context)
    assert list(graph.ready()) == [ta1, ta2]

    graph.set_status(ta1, TaskStatus.failed())
    assert list(graph.ready()) == [ta2]

    graph.set_status(ta2, TaskStatus.failed())
    assert list(graph.ready()) == [tb1]



from __future__ import annotations

import argparse
import builtins
import contextlib
import io
import json
import logging
import os
import pdb
import sys
import textwrap
from functools import partial
from itertools import chain
from pathlib import Path
from typing import Any, NoReturn

from kraken.common import (
    BuildscriptMetadata,
    CurrentDirectoryProjectFinder,
    LoggingOptions,
    RequirementSpec,
    appending_to_sys_path,
    get_terminal_width,
    not_none,
    propagate_argparse_formatter_to_subparser,
)
from kraken.common.pyenv import get_distributions
from nr.io.graphviz.render import render_to_browser
from nr.io.graphviz.writer import GraphvizWriter
from termcolor import colored

from kraken.core.address import Address, AddressResolutionError
from kraken.core.cli import serialize
from kraken.core.cli.executor import ColoredDefaultPrintingExecutorObserver, status_to_text
from kraken.core.cli.option_sets import BuildOptions, ExcludeOptions, GraphOptions, RunOptions, VizOptions
from kraken.core.system.context import Context
from kraken.core.system.errors import BuildError, ProjectNotFoundError
from kraken.core.system.graph import TaskGraph
from kraken.core.system.project import Project
from kraken.core.system.property import Property
from kraken.core.system.task import GroupTask, Task, TaskStatus

BUILD_SCRIPT = Path(".kraken.py")
BUILD_SUPPORT_DIRECTORY = "build-support"
logger = logging.getLogger(__name__)
print = partial(builtins.print, flush=True)


class BuildScriptError(Exception):
    """
    Raised if an exception occurs while executing the build script.
    """


def _get_argument_parser(prog: str) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog,
        formatter_class=lambda prog: argparse.RawDescriptionHelpFormatter(prog, width=120, max_help_position=60),
        description=textwrap.dedent(
            """
            The Kraken build system.

            Kraken focuses on ease of use and simplicity to model complex task orchestration workflows.
            """
        ),
    )
    subparsers = parser.add_subparsers(dest="cmd")

    run = subparsers.add_parser(
        "run",
        aliases=["r"],
        # The following makes sure --all will not be considered as an abbreviation of --allow-whatever
        # This is important because we do want to warn in case the user specifies the (invalid) --all flag
        # See https://docs.python.org/dev/library/argparse.html#allow-abbrev
        allow_abbrev=False,
    )
    LoggingOptions.add_to_parser(run)
    BuildOptions.add_to_parser(run)
    GraphOptions.add_to_parser(run)
    RunOptions.add_to_parser(run)

    query = subparsers.add_parser("query", aliases=["q"])
    query_subparsers = query.add_subparsers(dest="query_cmd")

    ls = query_subparsers.add_parser("ls", description="list all tasks and task groups in the build")
    LoggingOptions.add_to_parser(ls)
    BuildOptions.add_to_parser(ls)
    GraphOptions.add_to_parser(ls, saveable=False)

    describe = query_subparsers.add_parser(
        "describe",
        aliases=["d"],
        description="describe one or more tasks in detail",
    )
    LoggingOptions.add_to_parser(describe)
    BuildOptions.add_to_parser(describe)
    GraphOptions.add_to_parser(describe, saveable=False)

    viz = query_subparsers.add_parser("visualize", aliases=["viz", "v"], description="generate a GraphViz of the build")
    LoggingOptions.add_to_parser(viz)
    BuildOptions.add_to_parser(viz)
    GraphOptions.add_to_parser(viz, saveable=False)
    VizOptions.add_to_parser(viz)

    tree = query_subparsers.add_parser("tree", aliases=["t"], description="Output the project and task tree.")
    LoggingOptions.add_to_parser(tree)
    BuildOptions.add_to_parser(tree)
    GraphOptions.add_to_parser(tree, saveable=False)
    ExcludeOptions.add_to_parser(tree)

    # This command is used by kraken-wrapper to produce a lock file.
    env = query_subparsers.add_parser("env", description="produce a JSON file of the Python environment distributions")
    LoggingOptions.add_to_parser(env)

    propagate_argparse_formatter_to_subparser(parser)
    return parser


def _load_build_state(
    exit_stack: contextlib.ExitStack,
    build_options: BuildOptions,
    graph_options: GraphOptions,
) -> tuple[Context, TaskGraph]:
    """
    This function loads the build state for the current working directory; which involves either executing the
    Kraken build script or loading one or more state files from their serialized form on disk.
    """

    if graph_options.restart and not graph_options.resume:
        raise ValueError("the --restart option requires the --resume flag")

    # Calculate the main subproject based on the project directory.
    root_directory = build_options.project_dir.absolute().resolve()
    try:
        subproject_directory = Path.cwd().relative_to(root_directory)
    except ValueError:
        raise ValueError(
            f"-p,--project-dir must be a parent directory of {Path.cwd()}, not a sibling/subdirectory "
            f"(got: {build_options.project_dir})"
        )

    # For consistency, we always act as if Kraken was run from the project root directory.
    # Using the `subproject_directory`, we later fitler down which tasks are selected / how relative
    # task references on the CLI are resolved.
    os.chdir(root_directory)

    project_info = CurrentDirectoryProjectFinder.default().find_project(Path.cwd())
    if not project_info:
        # We are OKAY with resuming a build from serialized state files even if no build script exists in the
        # current working directory; this is a feature that is often useful for debugging purposes when you want
        # to inspect the final state of a build, like from CI.
        if not graph_options.resume:
            raise ValueError(f'no Kraken build script found in the directory "{build_options.project_dir}"')

    # Before we can deserialize the build state, we must add the additional paths to `sys.path` that are defined
    # in by the script using the buildscript() function, or for backwards compatibility, in the file header as
    # comments.

    # Note that if we are simply going to execute the build script (i.e. not deserializing from state files),
    # we can rely on the buildscript() call in the script to update `sys.path`; but if the deprecated file header
    # is used to define the pythonpath we still need to parse it explicitly.

    # When we resume a build from serialized state files, we do not execute the build script. Thus, in order
    # to ensure we add the correct paths to `sys.path`, we need to parse the extract the build metadata again.
    if project_info and graph_options.resume and project_info.runner.has_buildscript_call(project_info.script):
        with BuildscriptMetadata.capture() as future:
            project_info.runner.execute_script(project_info.script, {})
        assert future.done()
        requirements = RequirementSpec.from_metadata(future.result())
        exit_stack.enter_context(appending_to_sys_path(requirements.pythonpath))

    context: Context | None = None

    # Deserialize the build state from files in the build state directory (+ extra dirs) if that is what
    # the user requested.
    if graph_options.resume:
        context, graph = serialize.load_build_state([build_options.state_dir] + build_options.additional_state_dirs)
        if not graph:
            raise ValueError("cannot --resume without build state")
        if graph and graph_options.restart:
            graph.restart()
        assert context is not None

    # Otherwise, we need to execute the build script.
    else:
        if build_options.no_load_project:
            raise ValueError(
                "no existing build state was loaded; typically that would load the root project "
                "but --no-load-project was specified."
            )

        # Register a callback for when the buildscript calls the buildscript() method. Any requirements passed
        # to the function are already expected to have been handled with by the Kraken wrapper, but we need to
        # handle the additions to `sys.path` here.
        def _buildscript_metadata_callback(metadata: BuildscriptMetadata) -> None:
            requirements = RequirementSpec.from_metadata(metadata)
            exit_stack.enter_context(appending_to_sys_path(requirements.pythonpath))

        context = Context(build_options.build_dir)

        with BuildscriptMetadata.callback(_buildscript_metadata_callback):
            try:
                context.load_project(Path.cwd())
            except BaseException as exc:
                raise BuildScriptError(
                    "An unexpected error occurred while executing the build script. Please check "
                    "check the earlier log messages for more details."
                ) from exc
            context.finalize()
            graph = TaskGraph(context)

    assert graph is not None

    # Serialize the build graph, even on failure, at the end of the build.
    if not graph_options.no_save:
        exit_stack.callback(
            lambda: serialize.save_build_state(build_options.state_dir, build_options.state_name, not_none(graph))
        )

    # Find the project from which we'll resolve relative task references based on the original current working
    # directory relative to the project root directory.
    relative_address = Address(":" + ":".join(subproject_directory.parts))

    # Find the project that contains the current working directory.
    try:
        context.focus_project = context.get_project(relative_address)
        logger.info("project '%s' is the main project", context.focus_project.address)
    except ProjectNotFoundError as e:
        logger.info("project '%s' does not exist, setting context.focus_project = None", e.address)
        context.focus_project = None

    # Deselect all tasks.
    for task in graph.root.tasks():
        task.selected = False

    # Mark tasks that were explicitly selected on the command-line as such. Tasks may alter their behaviour
    # based on whether they were explicitly selected or not.
    if graph_options.tasks:
        selected = context.resolve_tasks(graph_options.tasks, relative_to=relative_address, set_selected=True)
        targets = selected
    else:
        targets = context.resolve_tasks(None, relative_to=relative_address)

    # Trim the graph down to the selected or default tasks.
    if not graph_options.all:
        graph = graph.trim(targets)

    return context, graph


def run(
    exit_stack: contextlib.ExitStack,
    build_options: BuildOptions,
    graph_options: GraphOptions,
    run_options: RunOptions,
) -> None:
    context, graph = _load_build_state(
        exit_stack=exit_stack,
        build_options=build_options,
        graph_options=graph_options,
    )

    context.observer = ColoredDefaultPrintingExecutorObserver()
    graph.mark_tasks_as_skipped(
        run_options.exclude_tasks or (),
        run_options.exclude_tasks_subgraph or (),
        reason="Excluded via CLI",
        origin="cli",
        reset=True,
    )

    if run_options.skip_build:
        print("note: skipped build due to -s,--skip-build option.")
        sys.exit(0)
    else:
        if not graph:
            if run_options.allow_no_tasks:
                print("note: no tasks were selected (--allow-no-tasks)", "blue", file=sys.stderr)
                sys.exit(0)
            else:
                print("error: no tasks were selected", file=sys.stderr)
                sys.exit(1)

        try:
            context.execute(graph)
        except BuildError as exc:
            print()
            print("error:", exc, file=sys.stderr)
            sys.exit(1)


def ls(graph: TaskGraph) -> None:
    goal_tasks = set(graph.tasks(goals=True))
    all_tasks = set(graph.tasks())
    if not all_tasks:
        print("no tasks")
        sys.exit(1)
    longest_name = max(map(len, (str(t.address) for t in all_tasks))) + 1

    print()
    print(colored("Tasks", "blue", attrs=["bold", "underline"]))
    print()

    width = get_terminal_width(120)

    def _print_task(task: Task) -> None:
        line = [str(task.address).ljust(longest_name)]
        remaining_width = width - len(line[0])
        if task in goal_tasks:
            line[0] = colored(line[0], "green")
        if task.default:
            line[0] = colored(line[0], attrs=["bold"])
        status = graph.get_status(task)
        if status is not None:
            line.append(f"[{status_to_text(status)}]")
            status_length = 2 + len(status_to_text(status, colored=False)) + 1
            remaining_width -= status_length
        description = task.get_description()
        if description:
            remaining_width -= 2
            if remaining_width <= 0:
                remaining_width = width
            parts = textwrap.wrap(description, remaining_width)
            line.append(parts[0])  # the first line is already indented
            line.append("\n")
            for part in parts[1:]:
                line.append((width - remaining_width - 1) * " ")
                line.append(part)
                line.append("\n")
            line.pop()
        print("  " + " ".join(line))

    def sort_key(task: Task) -> str:
        return str(task.address)

    for task in sorted(graph.tasks(), key=sort_key):
        if isinstance(task, GroupTask):
            continue
        _print_task(task)

    print()
    print(colored("Groups", "blue", attrs=["bold", "underline"]))
    print()

    for task in sorted(graph.tasks(), key=sort_key):
        if not isinstance(task, GroupTask):
            continue
        _print_task(task)

    print()


def tree(graph: TaskGraph, exclude_options: ExcludeOptions) -> None:
    if exclude_options.exclude_tasks or exclude_options.exclude_tasks_subgraph:
        # Mark tasks that are excluded as skipped.
        for task in chain(
            graph.context.resolve_tasks(exclude_options.exclude_tasks),
            graph.context.resolve_tasks(exclude_options.exclude_tasks_subgraph),
        ):
            if not graph.get_status(task):
                graph.set_status(task, TaskStatus.skipped("excluded by -x/-X"))

    tasks = set(graph.tasks())

    # Filter out empty group tasks.
    tasks = {t for t in tasks if not isinstance(t, GroupTask) or t.tasks}

    # Find the projects we'd be looking to output.
    projects = {x.project for x in tasks}
    if graph.context.focus_project:
        projects.add(graph.context.focus_project)

    # Make sure we include all projects in between.
    for p in list(projects):
        while p and p.parent:
            p = p.parent
            projects.add(p)

    def _format_address(obj: Project | Task) -> str:
        address = obj.address

        parent = address.parent.set_container(True) if address and not address.is_root() else None
        if parent:
            result = colored(str(parent), "grey")
        else:
            result = ""
        result = result + colored(":" if address.is_root() else address.name, attrs=["bold"])

        if isinstance(obj, Project):
            if address.is_root():
                result += colored(" (root project)", "green")
            else:
                result += colored(" (sub project)", "green")
        if obj == graph.context.focus_project:
            result += colored(" (focus)", "cyan")
        if isinstance(obj, GroupTask):
            if obj.default:
                result += colored(" (default group)", "yellow", attrs=["bold"])
            else:
                result += colored(" (group)", "yellow")
        elif isinstance(obj, Task):
            if obj.default:
                result += colored(" (default task)", "blue", attrs=["bold"])
            else:
                result += colored(" (task)", "blue")
        if isinstance(obj, Task):
            if obj.selected:
                result += colored(" (selected)", "magenta")

            status = graph.get_status(obj)
            if status is not None:
                result += colored(f" [{status_to_text(status)}]", "cyan")

        return result

    def _recurse(obj: Project | Task, prefix: str, is_last: bool) -> None:
        if not (isinstance(obj, Project) and obj.address.is_root()):
            if is_last:
                this_prefix = prefix + "└── "
                child_prefix = prefix + "    "
            else:
                this_prefix = prefix + "├── "
                child_prefix = prefix + "│   "
        else:
            this_prefix = child_prefix = ""
        if isinstance(obj, Project):
            print(f"{colored(this_prefix, 'grey')}{_format_address(obj)}")
            members = sorted(list(obj.tasks().values()) + list(obj.subprojects().values()), key=lambda x: x.name)
            # Don't show empty groups
            members = [m for m in members if not isinstance(m, GroupTask) or m.tasks]
            # Don't show tasks/projects that are not in the graph.
            members = [m for m in members if m in tasks or m in projects]
            for idx, member in enumerate(members):
                is_last = idx == len(members) - 1
                _recurse(member, child_prefix, is_last)
        else:
            print(f"{colored(this_prefix, 'grey')}{_format_address(obj)}")

    print()
    _recurse(graph.context.root_project, "", True)


def describe(graph: TaskGraph) -> None:
    """
    Describes the selected tasks.
    """

    tasks = [t for t in graph.tasks() if t.selected]
    print("selected", len(tasks), "task(s)")
    print()

    for task in tasks:
        print(
            "Group" if isinstance(task, GroupTask) else "Task", colored(str(task.address), attrs=["bold", "underline"])
        )
        print("  Type:", type(task).__module__ + "." + type(task).__name__)
        print("  Type defined in:", colored(sys.modules[type(task).__module__].__file__ or "???", "cyan"))
        print("  Default:", task.default)
        print("  Selected:", task.selected)
        # TODO(@NiklasRosenstein): Show task tags
        rels = list(task.get_relationships())
        print(colored("  Relationships", attrs=["bold"]), f"({len(rels)})")
        for rel in rels:
            print(
                "".ljust(4),
                colored(str(rel.other_task.address), "blue"),
                f"before={rel.inverse}, strict={rel.strict}",
            )
        print("  " + colored("Properties", attrs=["bold"]) + f" ({len(type(task).__schema__)})")
        longest_property_name = max(map(len, type(task).__schema__.keys())) if type(task).__schema__ else 0
        for key in type(task).__schema__:
            prop: Property[Any] = getattr(task, key)
            print(
                "".ljust(4),
                (key + ":").ljust(longest_property_name + 1),
                f'{colored(prop.get_or("<unset>"), "blue")}',
            )
        print()


def visualize(graph: TaskGraph, viz_options: VizOptions) -> None:
    root = graph.root
    if viz_options.reduce or viz_options.reduce_keep_explicit:
        root = root.reduce(keep_explicit=viz_options.reduce_keep_explicit)
        graph = graph.reduce(keep_explicit=viz_options.reduce_keep_explicit)

    buffer = io.StringIO()
    writer = GraphvizWriter(buffer if viz_options.show else sys.stdout)
    writer.digraph(fontname="monospace", rankdir="LR")
    writer.set_node_style(style="filled", shape="box")

    style_default = {"penwidth": "3"}
    style_goal = {"fillcolor": "lawngreen"}
    style_select = {"fillcolor": "darkgoldenrod1"}
    style_group = {"shape": "ellipse"}
    style_edge_non_strict = {"style": "dashed"}
    style_edge_implicit = {"color": "gray"}

    writer.subgraph("cluster_#legend", label="Legend")
    # writer.node("#task", label="task")
    writer.node("#group", label="group task", **style_group)
    writer.node("#default", label="runs by default", **style_default)
    writer.node("#selected", label="task will run", **style_select)
    writer.node("#goal", label="goal task", **style_goal)
    writer.end()

    writer.subgraph("cluster_#build", label="Build Graph")

    main = root if viz_options.inactive else graph
    goal_tasks = set(graph.tasks(goals=True))
    selected_tasks = set(graph.tasks())

    for task in main.tasks():
        style = {}
        style.update(style_default if task.default else {})
        style.update(style_group if isinstance(task, GroupTask) else {})
        style.update(style_select if task in selected_tasks else {})
        style.update(style_goal if task in goal_tasks else {})

        writer.node(str(task.address), **style)
        for predecessor in main.get_predecessors(task, ignore_groups=False):
            writer.edge(
                str(predecessor.address),
                str(task.address),
                **({} if main.get_edge(predecessor, task).strict else style_edge_non_strict),
                **(style_edge_implicit if main.get_edge(predecessor, task).implicit else {}),
            )

    writer.end()
    writer.end()

    if viz_options.show:
        render_to_browser(buffer.getvalue())


def env() -> None:
    dists = sorted(get_distributions().values(), key=lambda dist: dist.name)
    print(json.dumps([dist.to_json() for dist in dists], sort_keys=True))


def on_exception(exc: BaseException) -> int:
    """
    Called when an exception occurrs in #main_internal() to handle common errors and provide better error messages.
    """

    issues_url = "https://github.com/kraken-build/kraken-build/issues"

    # The wrapper will set `KRAKENW=1` so we can detect if we're running in the wrapper or not.
    command = "krakenw" if os.getenv("KRAKENW") else "kraken"

    match exc:
        case SystemExit():
            if not isinstance(exc.code, int):
                logger.warning("SystemExit.code is not an integer: %r", exc.code)
                return 1
            return exc.code
        case AddressResolutionError():
            logger.error(
                "No task matched the selector '%s'. Not finding what you're looking for? You can use the "
                "'%s query tree --all' command to list every available task.",
                exc.query,
                command,
            )
            return 1
        case BuildScriptError():
            logger.error(
                "An unexpected error occurred when executing the build script. This either indicates a mistake "
                "in your build script(s), or a bug in Kraken. If you think this is a bug in Kraken, please "
                "consider filing a bug report in the respective Kraken component repository. (If the traceback points "
                "to a bug in the core components, please file the report at %s)\n\n",
                issues_url,
                exc_info=exc.__context__ or exc,
            )
            return 2
        case _:
            logger.error(
                "An unexpected error occurred in the Kraken CLI. This is likely a bug in Kraken. Please consider "
                "filing a bug report at %s.\n\n",
                issues_url,
                exc_info=exc,
            )
            return 3


def main_internal(prog: str, argv: list[str] | None, pdb_enabled: bool) -> NoReturn:
    parser = _get_argument_parser(prog)
    args = parser.parse_args(sys.argv[1:] if argv is None else argv)

    if not args.cmd:
        parser.print_usage()
        sys.exit(0)

    if LoggingOptions.available(args):
        LoggingOptions.collect(args).init_logging()

    if pdb_enabled:
        logger.info("note: KRAKEN_PDB=1 is set, an interactive debugging session will be started on exit.")
        logger.info("note: Exceptions raised in tasks will not trigger an interactive debugging session.")

    if args.cmd in ("run", "r"):
        with contextlib.ExitStack() as exit_stack:
            run(exit_stack, BuildOptions.collect(args), GraphOptions.collect(args), RunOptions.collect(args))

    elif args.cmd in ("query", "q"):
        if not args.query_cmd:
            parser.print_usage()
            sys.exit(0)

        if args.query_cmd == "env":
            env()
            sys.exit(0)

        build_options = BuildOptions.collect(args)
        graph_options = GraphOptions.collect(args)

        with contextlib.ExitStack() as exit_stack:
            _context, graph = _load_build_state(
                exit_stack=exit_stack,
                build_options=build_options,
                graph_options=graph_options,
            )

            if args.query_cmd == "ls":
                ls(graph)
            elif args.query_cmd in ("describe", "d"):
                describe(graph)
            elif args.query_cmd in ("visualize", "viz", "v"):
                visualize(graph, VizOptions.collect(args))
            elif args.query_cmd in ("t", "tree"):
                tree(graph, ExcludeOptions.collect(args))
            else:
                assert False, args.query_cmd

    else:
        parser.print_usage()

    sys.exit(0)


def main(prog: str = "kraken", argv: list[str] | None = None, handle_exceptions: bool = True) -> NoReturn:
    pdb_enabled = os.getenv("KRAKEN_PDB") == "1"
    profile_outfile = os.getenv("KRAKEN_PROFILING")
    try:
        if profile_outfile:
            import cProfile as profile

            with open(profile_outfile, "w"):  # Make sure the file exists
                pass

            prof = profile.Profile()
            try:
                prof.runcall(main_internal, prog, argv, pdb_enabled)
            finally:
                prof.dump_stats(profile_outfile)
        else:
            main_internal(prog, argv, pdb_enabled)
        sys.exit(0)
    except BaseException as exc:
        if not handle_exceptions:
            raise
        code = on_exception(exc)
        if pdb_enabled:
            pdb.post_mortem()
        sys.exit(code)


if __name__ == "__main__":
    main()



from typing import Generator

from _pytest.capture import CaptureFixture, CaptureResult
from pytest import raises

from kraken.core.system.executor import GraphExecutorObserver
from kraken.core.system.executor.default import (
    TASKS_SKIPPED_DUE_TO_FAILING_DEPENDENCIES_TITLE,
    DefaultGraphExecutor,
    DefaultPrintingExecutorObserver,
    DefaultTaskExecutor,
)
from kraken.core.system.graph import TaskGraph
from kraken.core.system.project import Project
from kraken.core.system.task import Task, TaskStatus, VoidTask


class MyTask(Task):
    """
    Fake task
    """

    def execute(self) -> None:
        print("Hello")


class MyFailingTask(Task):
    """
    Fake failing task
    """

    def execute(self) -> None:
        raise RuntimeError("Wow this is failing")


def execute_print_test(graph: TaskGraph) -> None:
    """
    Basic common execution to get the final printed result
    """
    default_task_executor = DefaultTaskExecutor()
    default_printing_executor_observer = DefaultPrintingExecutorObserver()
    default_executor = DefaultGraphExecutor(default_task_executor)
    default_executor.execute_graph(graph, default_printing_executor_observer)


def trim_printed_result(captured: CaptureResult[str]) -> str:
    """
    Retrieving the skipped test printed part
    """
    print_output = str(captured.out)
    title_index_end = print_output.find(TASKS_SKIPPED_DUE_TO_FAILING_DEPENDENCIES_TITLE) + len(
        TASKS_SKIPPED_DUE_TO_FAILING_DEPENDENCIES_TITLE
    )
    result = print_output[(title_index_end + 1) :]
    return result


def test__DefaultExecutor__print_correct_failures_with_dependencies(
    kraken_project: Project, capsys: Generator[CaptureFixture[str], None, None]
) -> None:
    """This test tests if when a task failed, successor tasks with depedencies will be printed as failed.

    ```
    A -> B -> C -> D
    ```

    If B fails, C, D should be printed.
    """
    task_a = kraken_project.task("fake_task_a", MyTask)
    task_b = kraken_project.task("fake_task_b", MyFailingTask)
    task_c = kraken_project.task("fake_task_c", MyTask)
    task_d = kraken_project.task("fake_task_d", MyTask)

    task_b.depends_on(task_a)
    task_c.depends_on(task_b)
    task_d.depends_on(task_c)

    graph = TaskGraph(kraken_project.context).trim([task_d])
    assert set(graph.tasks()) == {task_a, task_b, task_c, task_d}
    execute_print_test(graph)
    captured = capsys.readouterr()  # type: ignore
    result = trim_printed_result(captured)
    assert TASKS_SKIPPED_DUE_TO_FAILING_DEPENDENCIES_TITLE in captured.out
    assert ":fake_task_c" in result
    assert ":fake_task_d" in result


def test__DefaultExecutor__print_correct_failures_inside_groups_without_dependencies(
    kraken_project: Project, capsys: Generator[CaptureFixture[str], None, None]
) -> None:
    """This test tests if when a task failed within a group, the group will not be printed as failed.

    ```
    Group "group", Tasks: A, B, C
    ```

    If B fails, C is not printed as failed.
    """

    kraken_project.task("fake_task_a", MyTask, group="group")
    kraken_project.task("fake_task_b", MyFailingTask, group="group")
    kraken_project.task("fake_task_c", MyTask, group="group")

    group = kraken_project.group("group")

    graph = TaskGraph(kraken_project.context).trim([group])
    execute_print_test(graph)
    captured = capsys.readouterr()  # type: ignore
    assert TASKS_SKIPPED_DUE_TO_FAILING_DEPENDENCIES_TITLE not in captured.out
    result = trim_printed_result(captured)
    assert ":group" not in result


def test__DefaultExecutor__print_correct_failures_inside_group_with_dependency(
    kraken_project: Project, capsys: Generator[CaptureFixture[str], None, None]
) -> None:
    """This test tests if when a task failed within a group,
     successor tasks with dependencies will be printed as failed.

    ```
    Group
    A
    |
    v
    B
    | \
    v  v
    C  D
    ```
    If B fails, C and D should be printed as failed
    """
    task_a = kraken_project.task("fake_task_a", MyTask, group="group")
    task_b = kraken_project.task("fake_task_b", MyFailingTask, group="group")
    task_c = kraken_project.task("fake_task_c", MyTask, group="group")
    task_d = kraken_project.task("fake_task_d", MyTask, group="group")

    group = kraken_project.group("group")

    task_b.depends_on(task_a)
    task_c.depends_on(task_b)
    task_d.depends_on(task_b)

    graph = TaskGraph(kraken_project.context).trim([group])
    execute_print_test(graph)
    captured = capsys.readouterr()  # type: ignore
    assert TASKS_SKIPPED_DUE_TO_FAILING_DEPENDENCIES_TITLE in captured.out
    result = trim_printed_result(captured)
    assert ":fake_task_c" in result
    assert ":fake_task_d" in result
    assert ":fake_task_b" not in result


def test__DefaultExecutor__print_correct_failures_with_independent_groups(
    kraken_project: Project, capsys: Generator[CaptureFixture[str], None, None]
) -> None:
    """This test tests if when a task failed within one group, a following independent group will not be affected
    ```
    Group 1,  Group 2
    A           D
    |           |
    v           v
    B           E
    |           |
    v           v
    C           F
    ```
    If B fails, C will be printed
    """
    task_a = kraken_project.task("fake_task_a", MyTask, group="g1")
    task_b = kraken_project.task("fake_task_b", MyFailingTask, group="g1")
    task_c = kraken_project.task("fake_task_c", MyTask, group="g1")
    task_d = kraken_project.task("fake_task_d", MyTask, group="g2")
    task_e = kraken_project.task("fake_task_e", MyTask, group="g2")
    task_f = kraken_project.task("fake_task_f", MyTask, group="g2")

    g1 = kraken_project.group("g1")
    g2 = kraken_project.group("g2")

    task_b.depends_on(task_a)
    task_c.depends_on(task_b)
    task_e.depends_on(task_d)
    task_f.depends_on(task_e)

    graph = TaskGraph(kraken_project.context).trim([g1, g2])
    execute_print_test(graph)
    captured = capsys.readouterr()  # type: ignore
    assert TASKS_SKIPPED_DUE_TO_FAILING_DEPENDENCIES_TITLE in captured.out
    result = trim_printed_result(captured)
    assert ":fake_task_c" in result
    assert ":fake_task_d" not in result
    assert ":fake_task_e" not in result
    assert ":fake_task_f" not in result


def test__DefaultExecutor__print_correct_failures_with_dependent_groups(
    kraken_project: Project, capsys: Generator[CaptureFixture[str], None, None]
) -> None:
    """This test tests if when a task failed within one group,
     successor tasks with dependencies will be printed as failed.
    If another group has a relationship with the failing group, its tasks will also be printed as failed.
    ```
    Group 1 --> Group 2
    A           D
    |           |
    v           v
    B           E
    |           |
    v           v
    C           F
    ```
    If B fails, C, D, E and F will be printed
    """
    task_a = kraken_project.task("fake_task_a", MyTask, group="g1")
    task_b = kraken_project.task("fake_task_b", MyFailingTask, group="g1")
    task_c = kraken_project.task("fake_task_c", MyTask, group="g1")
    task_d = kraken_project.task("fake_task_d", MyTask, group="g2")
    task_e = kraken_project.task("fake_task_e", MyTask, group="g2")
    task_f = kraken_project.task("fake_task_f", MyTask, group="g2")

    g1 = kraken_project.group("g1")
    g2 = kraken_project.group("g2")

    task_b.depends_on(task_a)
    task_c.depends_on(task_b)
    task_e.depends_on(task_d)
    task_f.depends_on(task_e)

    g2.depends_on(g1)

    graph = TaskGraph(kraken_project.context).trim([g2])
    execute_print_test(graph)
    captured = capsys.readouterr()  # type: ignore
    assert TASKS_SKIPPED_DUE_TO_FAILING_DEPENDENCIES_TITLE in captured.out
    result = trim_printed_result(captured)
    assert ":fake_task_c" in result
    assert ":fake_task_d" in result
    assert ":fake_task_e" in result
    assert ":fake_task_f" in result


def test__DefaultTaskExecutor__skips_tasks_to_be_skipped(kraken_project: Project) -> None:
    t1 = kraken_project.task("t1", VoidTask)
    t1.add_tag("skip", reason="This task must be skipped.")
    t2 = kraken_project.task("t2", MyTask)

    statuses: list[TaskStatus] = []

    def done(status: TaskStatus) -> None:
        statuses.append(status)

    with raises(RuntimeError) as excinfo:
        DefaultTaskExecutor().execute_task(t1, done)
    assert str(excinfo.value) == f"Tasks that are set to be skipped must not be passed into the task executor: {t1!r}"
    DefaultTaskExecutor().execute_task(t2, done)

    assert statuses == [TaskStatus.succeeded()]


def test__DefaultGraphExecutor__skips_tasks_to_be_skipped(kraken_project: Project) -> None:
    t1 = kraken_project.task("t1", VoidTask)
    t1.add_tag("skip", reason="This task must be skipped.")
    kraken_project.task("t2", MyTask)

    statuses: list[TaskStatus] = []

    class Observer(GraphExecutorObserver):
        def after_execute_task(self, task: Task, status: TaskStatus) -> None:
            statuses.append(status)

    executor = DefaultGraphExecutor(DefaultTaskExecutor())
    executor.execute_graph(TaskGraph(kraken_project.context), Observer())

    assert statuses == [TaskStatus.skipped("This task must be skipped."), TaskStatus.succeeded()]



"""
This module provides the property API which is used to describe properties on Kraken objects.

Properties are used to represent data on an object that is evaluated lazily. A property value can be filled
with static data or derived from other properties. When a property has no value set, it raises a
:class:`Supplier.Empty` exception (also accessible as :attr:`Property.Empty`).

A property may be marked as an "output" property, which means that its value is only known at a later stage
when some work has been performed (e.g. the owning Kraken task was executed). If such a property is evaluated
before the value is available, a :class:`Property.Deferred` exception is raised instead.
"""

from __future__ import annotations

import copy
import dataclasses
import sys
import weakref
from operator import concat
from pathlib import Path
from typing import Any, Callable, ClassVar, Iterable, Mapping, Sequence, TypeVar, cast

import deprecated
from kraken.common import NotSet, Supplier, not_none
from typeapi import (
    AnnotatedTypeHint,
    ClassTypeHint,
    LiteralTypeHint,
    TupleTypeHint,
    TypeHint,
    UnionTypeHint,
    get_annotations,
)

T = TypeVar("T")
U = TypeVar("U")


@dataclasses.dataclass
class PropertyConfig:
    """Used to annotate properties, to configure the property.

    .. code:: Example

        from kraken.core.system.property import Object, Property, output
        from typing_extensions import Annotated

        class MyObj(Object):
            a: Annotated[Property[int], PropertyConfig(output=True)]
    """

    output: bool = False
    default: Any | NotSet = NotSet.Value
    default_factory: Callable[[], Any] | NotSet = NotSet.Value
    help: str | None = None


@dataclasses.dataclass
class PropertyDescriptor:
    name: str
    is_output: bool
    default: Any | NotSet
    default_factory: Callable[[], Any] | NotSet
    item_type: TypeHint
    help: str | None = None

    def has_default(self) -> bool:
        return not (self.default is NotSet.Value and self.default_factory is NotSet.Value)

    def get_default(self) -> Any:
        if self.default is not NotSet.Value:
            return copy.deepcopy(self.default)
        elif self.default_factory is not NotSet.Value:
            return self.default_factory()
        else:
            raise RuntimeError(f"property {self.name!r} has no default value")


class Property(Supplier[T]):
    """A property represents an input or output parameter of an :class:`Object`."""

    class Deferred(Exception):
        """
        This exception is raised when an output property has no value set. It is distinct from the
        :class:`Supplier.Empty` exception in that it will propagate to the caller in any case.
        """

        def __init__(self, property: Property[Any], message: str | None = None) -> None:
            self.property = property
            self.message = message

        def __str__(self) -> str:
            if self.message:
                return f"{self.message} ({self.property})"
            else:
                return f"the value of {self.property} will be known at a later time"

    ValueAdapter = Callable[[Any], Any]

    # This dictionary is a registry for type adapters that are used to ensure that values passed
    # into a property with :meth:`set()` are of the appropriate type. If a type adapter for a
    # particular type does not exist, a basic type check is performed. Note that the type adaptation
    # is not particularly sophisticated at this point and will not apply on items in nested structures.
    VALUE_ADAPTERS: ClassVar[dict[type, ValueAdapter]] = {}

    @staticmethod
    def output(*, help: str | None = None) -> Any:
        """Assign the result of this function as a default value to a property on the class level of an :class:`Object`
        subclass to mark it as an output property. This is an alternative to using the :class:`typing.Annotated` type
        hint.

        .. code:: Example

            from kraken.core.system.property import Object, Property, output

            class MyObj(Object):
                a: Property[int] = output()
        """

        return PropertyConfig(output=True, help=help)

    @staticmethod
    def required(*, help: str | None = None) -> Any:
        """
        Assign the result of this function as a default value to a property class to declare that it is required. This
        is the default behaviour of the a property, so this function is only useful to specify a help text or to make
        it more explicit in the code.
        """

        return PropertyConfig(help=help)

    @staticmethod
    def default(value: Any, *, help: str | None = None) -> Any:
        """Assign the result of this function as a default value to a property to declare it's default value."""

        return PropertyConfig(default=value, help=help)

    @staticmethod
    def default_factory(func: Callable[[], Any], help: str | None = None) -> Any:
        """Assign the result of this function as a default value to a property to declare it's default factory."""

        return PropertyConfig(default_factory=func, help=help)

    @staticmethod
    @deprecated.deprecated(reason="use Property.required(), .default() or .default_factory() instead")
    def config(
        output: bool = False,
        default: Any | NotSet = NotSet.Value,
        default_factory: Callable[[], Any] | NotSet = NotSet.Value,
    ) -> Any:
        """Assign the result of this function as a default value to a property on the class level of an :class:`Object`
        subclass to configure it's default value or whether it is an output property. This is an alternative to using
        a :class:`typing.Annotated` type hint.

        .. code:: Example

            from kraken.core.system.property import Object, Property, config

            class MyObj(Object):
                a: Property[int] = config(default=42)
        """

        return PropertyConfig(output, default, default_factory)

    def __init__(
        self,
        owner: PropertyContainer | type[PropertyContainer],
        name: str,
        item_type: TypeHint | Any,
        deferred: bool = False,
        help: str | None = None,
    ) -> None:
        """
        :param owner: The object that owns the property instance.
        :param name: The name of the property.
        :param item_type: The original inner type hint of the property (excluding the Property type itself).
        :param deferred: Whether the property should be initialized with a :class:`DeferredSupplier`.
        :param help: A help text for the property.
        """

        # NOTE(@NiklasRosenstein): We expect that any union member be a ClassTypeHint or TupleTypeHint.
        def _get_types(hint: TypeHint) -> tuple[type, ...]:
            if isinstance(hint, (ClassTypeHint, TupleTypeHint)):
                return (hint.type,)
            elif isinstance(hint, LiteralTypeHint):
                # TODO(@NiklasRosenstein): Add validation to the property to error if a bad value is set.
                return tuple({type(x) for x in hint.values})
            else:
                raise RuntimeError(f"unexpected Property type hint {hint!r}")

        # Determine the accepted types of the property.
        item_type = item_type if isinstance(item_type, TypeHint) else TypeHint(item_type)
        if isinstance(item_type, UnionTypeHint):
            accepted_types = tuple(concat(*map(_get_types, item_type)))
        else:
            accepted_types = _get_types(item_type)

        # Ensure that we have value adapters for every accepted type.
        for accepted_type in accepted_types:
            if accepted_type not in self.VALUE_ADAPTERS:
                if not isinstance(accepted_type, type):
                    raise ValueError(f"missing value adapter for type {accepted_type!r}")
        assert len(accepted_types) > 0

        self.owner = owner
        self.name = name
        self.help = help
        self.accepted_types = accepted_types
        self.item_type = item_type
        self._value: Supplier[T] = DeferredSupplier(self) if deferred else Supplier.void()
        self._derived_from: Sequence[Supplier[Any]] = ()
        self._finalized = False
        self._error_message: str | None = None

    def __repr__(self) -> str:
        try:
            owner_fmt = str(self.owner)
        except Exception:
            owner_fmt = type(self.owner).__name__ + "(<exception during fmt>)"
        return f"Property({owner_fmt}.{self.name})"

    def _adapt_value(self, value: Any) -> Any:
        errors = []
        for accepted_type in self.accepted_types:
            try:
                adapter = self.VALUE_ADAPTERS[accepted_type]
            except KeyError:
                if isinstance(accepted_type, type):
                    adapter = _type_checking_adapter(accepted_type)
                else:
                    raise
            try:
                return adapter(value)
            except TypeError as exc:
                errors.append(exc)
        raise TypeError(f"{self}: " + "\n".join(map(str, errors))) from (errors[0] if len(errors) == 1 else None)

    @property
    def value(self) -> Supplier[T]:
        return self._value

    def derived_from(self) -> Iterable[Supplier[Any]]:
        yield self._value
        yield from self._value.derived_from()
        yield from self._derived_from

    def get(self) -> T:
        try:
            return self._value.get()
        except Supplier.Empty:
            raise Supplier.Empty(self, self._error_message)

    def set(self, value: T | Supplier[T], derived_from: Iterable[Supplier[Any]] = ()) -> None:
        if self._finalized:
            raise RuntimeError(f"{self} is finalized")
        derived_from = list(derived_from)
        if not isinstance(value, Supplier):
            value = Supplier.of(self._adapt_value(value), derived_from)
            derived_from = ()
        self._value = value
        self._derived_from = derived_from

    def setcallable(self, func: Callable[[], T], derived_from: Iterable[Supplier[Any]] = ()) -> None:
        if self._finalized:
            raise RuntimeError(f"{self} is finalized")
        if not callable(func):
            raise TypeError('"func" must be callable')
        self._value = Supplier.of_callable(func, list(derived_from))
        self._derived_from = ()

    def setmap(self, func: Callable[[T], T]) -> None:
        if self._finalized:
            raise RuntimeError(f"{self} is finalized")
        if not callable(func):
            raise TypeError('"func" must be callable')
        self._value = self._value.map(func)

    def setdefault(self, value: T | Supplier[T]) -> None:
        if self._finalized:
            raise RuntimeError(f"{self} is finalized")
        if self._value.is_void():
            self.set(value)

    def setfinal(self, value: T | Supplier[T]) -> None:
        self.set(value)
        self.finalize()

    def seterror(self, message: str) -> None:
        """Set an error message that should be included when the property is read."""

        self._error_message = message

    def clear(self) -> None:
        self.set(Supplier.void())

    def finalize(self) -> None:
        """Prevent further modification of the value in the property."""

        if not self._finalized:
            self._finalized = True

    def provides(self, type_: type) -> bool:
        """Returns `True` if the property may provide an instance or a sequence of the given *type_*."""

        if isinstance(self.item_type, UnionTypeHint):
            types = list(self.item_type)
        elif isinstance(self.item_type, ClassTypeHint):
            types = [self.item_type]
        else:
            assert False, self.item_type

        for provided in types:
            if not isinstance(provided, ClassTypeHint):
                continue
            if issubclass(provided.type, type_):
                return True
            if issubclass(provided.type, Sequence) and provided.args and len(provided.args) == 1:
                inner = provided.args[0]
                if issubclass(inner, type_):
                    return True

        return False

    def get_of_type(self, type_: type[U]) -> list[U]:
        """Return the inner value or values of the property as a flat list of *t*. If the property returns only a
        a single value of the specified type, the returned list will contain only that value. If the property instead
        provides a sequence that contains one or more objects of the provided type, only those objects will be
        returned.

        Note that this does not work with generic parametrized types."""

        value = self.get()
        if type_ != object and isinstance(value, type_):
            return [value]
        if isinstance(value, Sequence):
            return [x for x in value if isinstance(x, type_)]
        if type_ == object:
            return [cast(U, value)]
        return []

    @staticmethod
    def value_adapter(type_: type) -> Callable[[ValueAdapter], ValueAdapter]:
        """Decorator for functions that serve as a value adapter for the given *type_*."""

        def decorator(func: Property.ValueAdapter) -> Property.ValueAdapter:
            Property.VALUE_ADAPTERS[type_] = func
            return func

        return decorator

    def is_set(self) -> bool:
        """
        Returns #True if the property has been set to a value, #False otherwise. This is different from #is_empty(),
        because it does not require evaluation of the property value. This method reflects whether #set() has been
        called with any other value than a #VoidSupplier or a #DeferredSupplier.
        """

        return not self._value.is_void()

    # Supplier

    def is_empty(self) -> bool:
        if isinstance(self._value, DeferredSupplier):
            return True
        return super().is_empty()

    # Python Descriptor

    def __set__(self, instance: PropertyContainer, value: T | Supplier[T] | None) -> None:
        instance_prop = vars(instance)[self.name]
        assert isinstance(instance_prop, Property)
        if value is not None or type(None) in self.accepted_types:
            instance_prop.set(value)
        else:
            instance_prop.clear()

    def __get__(self, instance: PropertyContainer | None, owner: type[Any]) -> Property[T]:
        if instance is None:
            return self
        instance_prop = vars(instance)[self.name]
        assert isinstance(instance_prop, Property)
        return instance_prop


class PropertyContainer:
    """Base class. An object's schema is declared as annotations linking to properties."""

    __schema__: ClassVar[Mapping[str, PropertyDescriptor]] = {}

    def __init_subclass__(cls) -> None:
        """Initializes the :attr:`__schema__` by introspecting the class annotations."""

        schema: dict[str, PropertyDescriptor] = {}
        base: type[PropertyContainer]
        for base in cls.__bases__:
            if issubclass(base, PropertyContainer):
                schema.update(base.__schema__)

        for key, hint in get_annotations(cls).items():
            hint = TypeHint(hint)
            config: PropertyConfig | None = None

            # Unwrap annotatations, looking for a PropertyConfig annotation.
            if isinstance(hint, AnnotatedTypeHint):
                config = next((x for x in hint.metadata if isinstance(x, PropertyConfig)), None)
                hint = TypeHint(hint.type)

            # Check if :func:`output()` or :func:`default()` was used to configure the property.
            if hasattr(cls, key) and isinstance(getattr(cls, key), PropertyConfig):
                assert config is None, "PropertyConfig cannot be on both an attribute and type annotation"
                config = getattr(cls, key)
                delattr(cls, key)

            # Is the hint pointing to a Property type?
            if isinstance(hint, ClassTypeHint) and hint.type == Property:
                assert isinstance(hint, ClassTypeHint) and hint.type == Property, hint
                assert hint.args is not None and len(hint.args) == 1, hint
                config = config or PropertyConfig()
                schema[key] = PropertyDescriptor(
                    name=key,
                    is_output=config.output,
                    default=config.default,
                    default_factory=config.default_factory,
                    item_type=hint[0].evaluate(vars(sys.modules[cls.__module__])),
                    help=config.help,
                )

            # The attribute is annotated as an output but not actually typed as a property?
            elif config:
                raise RuntimeError(
                    f"Type hint for {cls.__name__}.{key} is annotated as a 'PropertyConfig', but not actually "
                    "typed as a 'Property'."
                )

            cls.__schema__ = schema

        # Make sure there's a Property descriptor on the class for every property in the schema.
        for key, value in cls.__schema__.items():
            setattr(cls, key, Property[Any](cls, key, value.item_type, value.is_output, value.help))

    def __init__(self) -> None:
        """Creates :class:`Properties <Property>` for every property defined in the object's schema."""

        for key, desc in self.__schema__.items():
            prop = Property[Any](self, key, desc.item_type, desc.is_output, desc.help)
            vars(self)[key] = prop
            if desc.has_default():
                prop.setdefault(desc.get_default())


class DeferredSupplier(Supplier[Any]):
    def __init__(self, property: Property[Any]) -> None:
        self.property = weakref.ref(property)

    def derived_from(self) -> Iterable[Supplier[Any]]:
        return ()

    def get(self) -> Any:
        raise Property.Deferred(not_none(self.property()))

    def is_void(self) -> bool:
        return True


# Register common value adapters


def _type_checking_adapter(type_: type) -> Property.ValueAdapter:
    def func(value: Any) -> Any:
        if not isinstance(value, type_):
            raise TypeError(f"expected {type_.__name__}, got {type(value).__name__}")
        return value

    func.__name__ = f"check_{type_.__name__}"
    return func


@Property.value_adapter(Path)
def _adapt_path(value: Any) -> Path:
    if isinstance(value, str):
        return Path(value)
    if not isinstance(value, Path):
        raise TypeError(f"expected Path, got {type(value).__name__}")
    return value



from __future__ import annotations

import re
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Mapping, TypeVar, cast, overload

import builddsl
from deprecated import deprecated
from typing_extensions import Literal

from kraken.core.address import Address
from kraken.core.base import Currentable, MetadataContainer
from kraken.core.system.kraken_object import KrakenObject
from kraken.core.system.property import Property
from kraken.core.system.task import GroupTask, InlineTask, Task, TaskSet

if TYPE_CHECKING:
    from kraken.core.system.context import Context

T = TypeVar("T")
T_Task = TypeVar("T_Task", bound="Task")


class TaskNotFound(Exception):
    pass


class DuplicateMember(Exception):
    pass


class Project(KrakenObject, MetadataContainer, Currentable["Project"]):
    """A project consolidates tasks related to a directory on the filesystem."""

    directory: Path
    context: Context
    metadata: list[Any]  #: A list of arbitrary objects that are usually looked up by type.

    def __init__(self, name: str, directory: Path, parent: Project | None, context: Context) -> None:
        assert isinstance(name, str), type(name)
        assert isinstance(directory, Path), type(directory)
        assert isinstance(parent, Project) or parent is None, type(parent)
        KrakenObject.__init__(self, name, parent)
        MetadataContainer.__init__(self)
        Currentable.__init__(self)

        self.directory = directory
        self.context = context
        self.metadata = []

        # We store all members that can be referenced by a fully qualified name in the same dictionary to ensure
        # we're not accidentally allocating the same name twice.
        self._members: dict[str, Task | Project] = {}

        apply_group = self.group(
            "apply", description="Tasks that perform automatic updates to the project consistency."
        )
        fmt_group = self.group("fmt", description="Tasks that that perform code formatting operations.")
        fmt_group.depends_on(apply_group, mode="strict")

        check_group = self.group("check", description="Tasks that perform project consistency checks.", default=True)

        gen_group = self.group("gen", description="Tasks that perform code generation.", default=True)

        lint_group = self.group("lint", description="Tasks that perform code linting.", default=True)
        lint_group.depends_on(check_group, mode="strict")
        lint_group.depends_on(gen_group, mode="strict")

        build_group = self.group("build", description="Tasks that produce build artefacts.")
        build_group.depends_on(lint_group, mode="order-only")
        build_group.depends_on(gen_group, mode="strict")

        audit_group = self.group("audit", description="Tasks that perform auditing on built artefacts and code")
        audit_group.depends_on(build_group, mode="strict")
        audit_group.depends_on(gen_group, mode="strict")

        test_group = self.group("test", description="Tasks that perform unit tests.", default=True)
        test_group.depends_on(build_group, mode="order-only")
        test_group.depends_on(gen_group, mode="strict")

        integration_test_group = self.group("integrationTest", description="Tasks that perform integration tests.")
        integration_test_group.depends_on(test_group, mode="order-only")
        integration_test_group.depends_on(gen_group, mode="strict")

        publish_group = self.group("publish", description="Tasks that publish build artefacts.")
        publish_group.depends_on(integration_test_group, mode="order-only")
        publish_group.depends_on(build_group, mode="strict")

        deploy_group = self.group("deploy", description="Tasks that deploy applications.")
        deploy_group.depends_on(publish_group, mode="order-only")

        self.group("update", description="Tasks that update dependencies of the project.")

    def __repr__(self) -> str:
        return f"Project({self.address})"

    @property
    def parent(self) -> Project | None:
        if self._parent is None:
            return None
        assert isinstance(self._parent, Project), "Project.parent must be a Project"
        return self._parent

    @property
    def name(self) -> str:
        if self.address.is_root():
            warnings.warn(
                "Accessing Project.name on the root project is deprecated since kraken-core v0.12.0. "
                "In future versions, this will result ValueError being raised. The project name is now "
                "determined by the Address.name, which is undefined on the root address (`:`). "
                "The fallback behaviour for this version is that we return the Project.directory.name.",
                DeprecationWarning,
            )
            return self.directory.name
        return self.address.name

    @property
    def build_directory(self) -> Path:
        """Returns the recommended build directory for the project; this is a directory inside the context
        build directory ammended by the project name."""

        return self.context.build_directory / str(self.address).replace(":", "/").lstrip("/")

    @overload
    def task(self, name: str, /) -> Task:
        """
        Get a task from the project by name. Raises a #TaskNotFound exception if the task does not exist.
        """

    @overload
    def task(
        self,
        name: str,
        type_: type[T_Task],
        /,
        *,
        default: bool | None = None,
        group: str | GroupTask | None = None,
        description: str | None = None,
    ) -> T_Task:
        """
        Create a new task in the project with the specified name. If a member of the project with the same name already
        exists, a #DuplicateMember exception is raised.
        """

    @overload
    def task(self, name: str, type_: type[T_Task], closure: builddsl.UnboundClosure, /) -> T_Task:
        """
        This overload is used to create a task in a BuildDSL script.

        ```py
        project.task "myTask" MyTaskType {
            default = False
            some_property.set("foobar")
            depends_on "otherTask"
        }
        ```
        """

    @overload
    def task(
        self,
        name: str,
        closure: builddsl.UnboundClosure,
        /,
        *,
        default: bool | None = None,
        group: str | GroupTask | None = None,
        description: str | None = None,
    ) -> Task:
        """
        Create a new task, applying the *closure*. This is useful for BuildDSL scripts that want to implement
        a custom task. This can be done by overriding the task's `execute()` method with a closure. It will
        create a task of type #InlineTask for you.

        ```py
        project.task "myTask" {
            property "name" "John"
            execute = () -> {
                def name = self.property("name").get()
                print "Hello," name
            }
        }
        ```
        """

    def task(
        self,
        name: str,
        type_: type[T_Task] | builddsl.UnboundClosure | None = None,
        default_or_closure: bool | builddsl.UnboundClosure | None = None,
        /,
        *,
        default: bool | None = None,
        group: str | GroupTask | None = None,
        description: str | None = None,
    ) -> Task | T_Task:
        if type_ is None:
            assert default_or_closure is None
            assert default is None
            assert group is None
            assert description is None

            try:
                task = self._members[name]
                if not isinstance(task, Task):
                    raise KeyError("Not a task")
            except KeyError:
                raise TaskNotFound(self.address.concat(name))
            return task

        if isinstance(type_, builddsl.UnboundClosure):
            default_or_closure = type_
            type_ = InlineTask  # type: ignore[assignment]

        if type_ is None or not isinstance(type_, type) or not issubclass(type_, Task):
            raise TypeError(f"Expected a Task type, got {type(type_).__name__}")

        if name in self._members:
            raise DuplicateMember(f"{self} already has a member {name!r}")

        task = type_(name, self)

        if callable(default_or_closure):
            default_or_closure(task)
        elif default_or_closure is not None:
            assert isinstance(default_or_closure, bool)
            task.default = default_or_closure
        elif default is not None:
            task.default = default

        match group:
            case str():
                self.group(group).add(task)
            case GroupTask():
                group.add(task)
            case None:
                pass
            case _:
                raise TypeError(f"Expected str or GroupTask, got {type(group)}")

        if description is not None:
            task.description = description

        self._members[name] = task
        return task

    def tasks(self) -> Mapping[str, Task]:
        return {t.name: t for t in self._members.values() if isinstance(t, Task)}

    def subprojects(self) -> Mapping[str, Project]:
        return {p.name: p for p in self._members.values() if isinstance(p, Project)}

    @overload
    def subproject(self, name: str, mode: Literal["empty", "execute"] = "execute") -> Project:
        """
        Mount a sub-project of this project with the specified *name*.

        :param name: The name of the sub-project. The address of the returned project will be the current project's
            address appended by the given *name*. The name must not contain special characters reserved to the Address
            syntax. The sub-project will be bound to the directory with the same *name* in the directory of the current
            project.
        :param mode: Specifies how the project should be created. If set to "empty", the project will be created
            without loading any build scripts. If set to "execute" (default), the project will be created and its
            build scripts will be executed.
        """

    @overload
    def subproject(self, name: str, mode: Literal["if-exists"]) -> Project | None:
        """
        Mount a sub-project of this project with the specified *name* and execute it if the directory matching the
        *name* exists. If such a directory does not exist, no project is created and `None` is returned. If you want
        to create a project in any case, you can use this method, and if you get `None` back you can call
        #subproject() again with the *mode* set to "empty".
        """

    def subproject(
        self,
        name: str,
        mode: Literal["empty", "execute", "if-exists"] = "execute",
    ) -> Project | None:
        assert isinstance(mode, str), f"mode must be a string, got {type(mode).__name__}"

        obj = self._members.get(name)
        if obj is None and mode == "if-exists":
            return None
        if obj is not None:
            if not isinstance(obj, Project):
                raise ValueError(
                    f"{self.address}:{name} does not refer to a project (got {type(obj).__name__} instead)"
                )
            return obj

        directory = self.directory / name
        if mode == "empty":
            project = Project(name, directory, self, self.context)
            self._members[name] = project
        elif mode == "execute":
            if not directory.is_dir():
                raise FileNotFoundError(
                    f"{self.address}:{name} cannot be loaded because the directory {directory} does not exist"
                )
            project = self.context.load_project(directory, self, require_buildscript=False)
            assert name in self._members
            assert self._members[name] is project
        else:
            raise ValueError(f"invalid mode {mode!r}")

        return project

    def has_subproject(self, name: str) -> bool:
        """
        Returns `True` if *name* refers to a subproject that exists in the current project.
        """

        return isinstance(self._members.get(name), Project)

    def add_task(self, task: Task) -> None:
        """Adds a task to the project.

        Raises:
            ValueError: If a member with the same name already exists or if the task's project does not match
        """

        if task.name in self._members:
            raise ValueError(f"{self} already has a member {task.name!r}, cannot add {task}")
        if task.project is not self:
            raise ValueError(f"{task}.project mismatch")
        self._members[task.name] = task

    def add_child(self, project: Project) -> None:
        """Adds a project as a child project.

        Raises:
            ValueError: If a member with the same name already exists or if the project's parent does not match
        """

        if project.name in self._members:
            raise ValueError(f"{self} already has a member {project.name!r}, cannot add {project}")
        if project.parent is not self:
            raise ValueError(f"{project}.parent mismatch")
        self._members[project.name] = project

    def remove_child(self, project: Project) -> None:
        assert project.parent is self
        assert self._members[project.name] is project

        del self._members[project.name]

    @deprecated(reason="Use Project.task() instead")
    def do(
        self,
        name: str,
        task_type: type[T_Task] = cast(Any, Task),
        default: bool | builddsl.UnboundClosure | None = None,
        *,
        group: str | GroupTask | None = None,
        description: str | None = None,
        **kwargs: Any,
    ) -> T_Task:
        """Add a task to the project under the given name, executing the specified action.

        :param name: The name of the task to add.
        :param task_type: The type of task to add.
        :param default: Override :attr:`Task.default`, or a closure to invoke with the created task.
        :param group: Add the task to the given group in the project.
        :param kwargs: Any number of properties to set on the task. Unknown properties will be ignored
            with a warning log.
        :return: The created task.
        """

        # NOTE(NiklasRosenstein): In versions prior to kraken-core 0.12.0, we did not validate task names.
        #       Now, the #Address class performs the validation and is rather strict. In order to not fully
        #       break usage of this function with invalid names, we convert the name to a valid form instead
        #       and issue a warning. This behaviour shall be removed in kraken-core 0.14.0.

        if not re.match(Address.Element.VALIDATION_REGEX, name):
            new_name = re.sub(f"[^{Address.Element.VALID_CHARACTERS}]+", "-", name)
            warnings.warn(
                f"Task name `{name}` is invalid and will be normalized to `{new_name}`. Starting with "
                "kraken-core 0.12.0, Task names must follow a stricter naming convention subject to the "
                f"Address class' validation (must match /{Address.Element.VALIDATION_REGEX}/).",
                DeprecationWarning,
                stacklevel=2,
            )
            name = new_name

        if name in self._members:
            raise ValueError(f"{self} already has a member {name!r}")

        task = task_type(name, self)
        if default is not None and not isinstance(default, builddsl.UnboundClosure):
            task.default = default
        if description is not None:
            task.description = description

        invalid_keys = set()
        for key, value in kwargs.items():
            prop = getattr(task, key, None)
            if isinstance(prop, Property):
                if value is not None:
                    prop.set(value)
            else:
                invalid_keys.add(key)
        if invalid_keys:
            task.logger.warning(
                "properties %s cannot be set because they don't exist (task %s)", invalid_keys, task.address
            )

        if isinstance(default, builddsl.UnboundClosure):
            default(task)
        self.add_task(task)
        if isinstance(group, str):
            group = self.group(group)
        if group is not None:
            group.add(task)
        return task

    def group(self, name: str, *, description: str | None = None, default: bool | None = None) -> GroupTask:
        """Create or get a group of the given name. If a task with the given name already exists, it must refer
        to a task of type :class:`GroupTask`, otherwise a :class:`RuntimeError` is raised.

        :param name: The name of the group in the project.
        :param description: If specified, set the group's description.
        :param default: Whether the task group is run by default."""

        task = self.tasks().get(name)
        if task is None:
            task = self.task(name, GroupTask)
        elif not isinstance(task, GroupTask):
            raise RuntimeError(f"{task.address!r} must be a GroupTask, but got {type(task).__name__}")
        if description is not None:
            task.description = description
        if default is not None:
            task.default = default

        return task

    ##
    # Begin: Deprecated APIs
    ##

    @property
    @deprecated(reason="Project.path is deprecated, use str(Project.address) instead")
    def path(self) -> str:
        """Returns the path that uniquely identifies the project in the current build context."""

        return str(self.address)

    @deprecated(reason="Project.resolve_tasks() is deprecated, use Project.context.resolve_tasks() instead.")
    def resolve_tasks(self, tasks: str | Task | Iterable[str | Task]) -> TaskSet:
        """Resolve tasks relative to the current project."""

        if isinstance(tasks, (str, Task)):
            tasks = [tasks]

        result = TaskSet()
        for item in tasks:
            if isinstance(item, str):
                result.add(self.context.resolve_tasks([item], self), partition=item)
            else:
                result.add([item])

        return result



from __future__ import annotations

import dataclasses
import uuid
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import argparse

DEFAULT_BUILD_DIR = Path("build")
BUILD_STATE_DIR = ".kraken/buildenv"


@dataclasses.dataclass(frozen=True)
class BuildOptions:
    build_dir: Path
    project_dir: Path
    state_dir: Path
    additional_state_dirs: list[Path]
    no_load_project: bool
    state_name: str

    @staticmethod
    def add_to_parser(parser: argparse.ArgumentParser) -> None:
        group = parser.add_argument_group("build options")
        group.add_argument(
            "-b",
            "--build-dir",
            metavar="PATH",
            type=Path,
            default=DEFAULT_BUILD_DIR,
            help="the build directory to write to [default: %(default)s]",
        )
        group.add_argument(
            "-p",
            "--project-dir",
            metavar="PATH",
            type=Path,
            default=Path.cwd(),
            help="the root directory of the project. If this is specified, it should point to an existing directory "
            "that contains a build script and it must be the same or a parent of the current directory. When "
            "invoked with this option, task references are resolved relative to the Kraken project that is "
            "represented by the current working directory. (note: this option is automatically passed when using "
            "kraken-wrapper as it finds the respective project automatically).",
        )
        group.add_argument(
            "--state-name",
            metavar="NAME",
            help="specify a name for the generated state file; if not specified, a short random ID is used",
            default=str(uuid.uuid4())[:7],
        )
        group.add_argument(
            "--state-dir",
            metavar="PATH",
            type=Path,
            help=f"specify the main build state directory [default: ${{--build-dir}}/{BUILD_STATE_DIR}]",
        )
        group.add_argument(
            "--additional-state-dir",
            metavar="PATH",
            type=Path,
            help="specify an additional state directory to load build state from. can be specified multiple times",
        )
        group.add_argument(
            "--no-load-project",
            action="store_true",
            help="do not load the root project. this is only useful when loading an existing build state",
        )

    @classmethod
    def collect(cls, args: argparse.Namespace) -> BuildOptions:
        return cls(
            build_dir=args.build_dir,
            project_dir=args.project_dir,
            state_name=args.state_name,
            state_dir=args.state_dir or args.build_dir / BUILD_STATE_DIR,
            additional_state_dirs=args.additional_state_dir or [],
            no_load_project=args.no_load_project,
        )


@dataclasses.dataclass(frozen=True)
class GraphOptions:
    tasks: list[str] | None
    resume: bool
    restart: bool
    no_save: bool
    all: bool

    @staticmethod
    def add_to_parser(parser: argparse.ArgumentParser, saveable: bool = True) -> None:
        group = parser.add_argument_group("graph options")
        group.add_argument("--resume", action="store_true", help="load previous build state")
        group.add_argument(
            "--restart",
            choices=("all",),
            help="load previous build state, but discard existing results (requires --resume)",
        )

        group.add_argument("--all", action="store_true", help="include all tasks in the build graph")

        # --all may make no sense
        # e.g. we don't want `kraken run --all` to be possible.
        # See https://github.com/kraken-build/kraken-core/pull/37#discussion_r1171300400
        group.add_argument(
            "tasks",
            metavar="task",
            nargs="*",
            help="one or more tasks to include in the build graph and mark as selected. if not set, default "
            "tasks are included in the build graph but not selected.",
            default=[],
        )

        if saveable:
            group.add_argument("--no-save", action="store_true", help="do not save the new build state")

    @classmethod
    def collect(cls, args: argparse.Namespace) -> GraphOptions:
        return cls(
            tasks=args.tasks or None,
            resume=args.resume,
            restart=args.restart,
            no_save=getattr(args, "no_save", True),
            all=getattr(args, "all", False),
        )


@dataclasses.dataclass(frozen=True)
class ExcludeOptions:
    exclude_tasks: list[str] | None
    exclude_tasks_subgraph: list[str] | None

    _group_name = "exclude options"

    @classmethod
    def add_to_parser(cls, parser: argparse.ArgumentParser) -> argparse._ArgumentGroup:
        group = parser.add_argument_group(cls._group_name)
        group.add_argument("-x", "--exclude", metavar="TASK", action="append", help="exclude one or more tasks")
        group.add_argument(
            "-X",
            "--exclude-subgraph",
            action="append",
            metavar="TASK",
            help="exclude the entire subgraphs of one or more tasks",
        )
        return group

    @classmethod
    def collect(cls, args: argparse.Namespace) -> ExcludeOptions:
        return cls(
            exclude_tasks=args.exclude or [],
            exclude_tasks_subgraph=args.exclude_subgraph or [],
        )


@dataclasses.dataclass(frozen=True)
class RunOptions(ExcludeOptions):
    allow_no_tasks: bool
    skip_build: bool

    _group_name = "run options"

    @classmethod
    def add_to_parser(cls, parser: argparse.ArgumentParser) -> argparse._ArgumentGroup:
        group = super().add_to_parser(parser)
        group.add_argument("-s", "--skip-build", action="store_true", help="just load the project, do not build")
        group.add_argument("-0", "--allow-no-tasks", action="store_true", help="don't error if no tasks got selected")
        return group

    @classmethod
    def collect(cls, args: argparse.Namespace) -> RunOptions:
        return cls(
            skip_build=args.skip_build,
            allow_no_tasks=args.allow_no_tasks,
            exclude_tasks=args.exclude or [],
            exclude_tasks_subgraph=args.exclude_subgraph or [],
        )


@dataclasses.dataclass(frozen=True)
class VizOptions:
    inactive: bool
    show: bool
    reduce: bool
    reduce_keep_explicit: bool

    @staticmethod
    def add_to_parser(parser: argparse.ArgumentParser) -> None:
        group = parser.add_argument_group("visualization options")
        group.add_argument("-i", "--inactive", action="store_true", help="include inactive tasks in the graph")
        group.add_argument("-s", "--show", action="store_true", help="show the graph in the browser (requires dot)")
        group.add_argument("-R", "--reduce", action="store_true", help="fully transitively reduce the graph")
        group.add_argument(
            "-r",
            "--reduce-keep-explicit",
            action="store_true",
            help="transitively reduce the graph but keep explicit edges",
        )

    @classmethod
    def collect(cls, args: argparse.Namespace) -> VizOptions:
        return cls(
            inactive=args.inactive,
            show=args.show,
            reduce=args.reduce,
            reduce_keep_explicit=args.reduce_keep_explicit,
        )



from __future__ import annotations

from kraken.core.address import Address, Addressable


class KrakenObject(Addressable):
    """
    A Kraken object is an object that can be addressed in a Kraken context.
    """

    _name: str
    _parent: KrakenObject | None

    def __init__(self, name: str, parent: KrakenObject | None = None):
        assert isinstance(name, str), type(name)
        assert isinstance(parent, KrakenObject) or parent is None, type(parent)
        self._name = name
        self._parent = parent

        # Validate that the name is valid by getting the object's address.
        self.address

    @property
    def name(self) -> str:
        return self._name

    @property
    def address(self) -> Address:
        """
        Returns the full address of the object.
        """

        if self._parent is None:
            return Address.ROOT
        else:
            return self._parent.address.append(self._name)



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable

from pytest import raises

from kraken.core.address import Address, AddressResolutionError, AddressSpace, resolve_address


@dataclass(init=False, repr=False)
class Node:
    parent: Node | None
    address: Address
    children: dict[str, Node]

    def __init__(self, parent: Node | None, name: str) -> None:
        self.parent = parent
        self.address = parent.address.append(name) if parent else Address.ROOT
        self.children = {}
        if parent:
            assert name not in parent.children, self.address
            parent.children[name] = self

    def __repr__(self) -> str:
        return f"Node(address={self.address!r})"


class NodeAddressSpace(AddressSpace[Node]):
    def __init__(self, root: Node) -> None:
        self._root = root

    def get_root(self) -> Node:
        return self._root

    def get_parent(self, entity: Node) -> Node | None:
        return entity.parent

    def get_children(self, entity: Node) -> Iterable[Node]:
        return entity.children.values()


class ExampleNodeTree:
    """Builds an example node tree."""

    def __init__(self) -> None:
        self.root = Node(None, "<root>")
        self.a = Node(self.root, "a")
        self.aa = Node(self.a, "a")
        self.ab = Node(self.a, "b")
        self.aba = Node(self.ab, "a")
        self.abb = Node(self.ab, "b")
        self.abc = Node(self.ab, "c")
        self.d = Node(self.root, "d")


def test__resolve_address__can_resolve_concrete_addresses() -> None:
    tree = ExampleNodeTree()
    space = NodeAddressSpace(tree.root)

    with raises(ValueError) as excinfo:
        resolve_address(space, tree.root, Address(""))
    assert str(excinfo.value) == "An empty address query cannot be resolved"

    assert list(resolve_address(space, tree.root, Address(":")).matches()) == [tree.root]
    assert list(resolve_address(space, tree.root, Address(":a")).matches()) == [tree.a]
    assert list(resolve_address(space, tree.root, Address(":a:b")).matches()) == [tree.ab]
    assert list(resolve_address(space, tree.root, Address(":a:b:a")).matches()) == [tree.aba]
    assert list(resolve_address(space, tree.root, Address(":a:b:c")).matches()) == [tree.abc]
    assert list(resolve_address(space, tree.root, Address(":d")).matches()) == [tree.d]
    assert list(resolve_address(space, tree.root, Address("d")).matches()) == [tree.d]
    assert list(resolve_address(space, tree.a, Address("..")).matches()) == [tree.root]
    assert list(resolve_address(space, tree.aa, Address("..:b")).matches()) == [tree.ab]
    assert list(resolve_address(space, tree.root, Address(":**:a")).matches()) == [tree.a, tree.aa, tree.aba]
    assert list(resolve_address(space, tree.root, Address(":**")).matches()) == [
        tree.a,
        tree.aa,
        tree.ab,
        tree.aba,
        tree.abb,
        tree.abc,
        tree.d,
    ]
    assert list(resolve_address(space, tree.root, Address(":**:")).matches()) == [tree.a, tree.ab]


def test__resolve_address__does_not_resolve_missing_address_and_raises_error() -> None:
    tree = ExampleNodeTree()
    space = NodeAddressSpace(tree.root)

    with raises(AddressResolutionError) as excinfo:
        resolve_address(space, tree.root, Address("a:c"))
    assert excinfo.value == AddressResolutionError(space, tree.root, Address("a:c"), tree.a, Address("c"))
    assert str(excinfo.value) == (
        "Could not resolve address 'a:c' in context ':'. The failure occurred at address "
        "':a' trying to resolve the remainder 'c'. The address ':a:c' does not exist."
    )


def test__resolve_address__can_resolve_wildcards() -> None:
    tree = ExampleNodeTree()
    space = NodeAddressSpace(tree.root)

    assert list(resolve_address(space, tree.root, Address("*")).matches()) == [tree.a, tree.d]
    assert list(resolve_address(space, tree.root, Address(":a:b:*")).matches()) == [tree.aba, tree.abb, tree.abc]


def test__resolve_address__can_resolve_recursive_wildcards() -> None:
    tree = ExampleNodeTree()
    space = NodeAddressSpace(tree.root)

    assert list(resolve_address(space, tree.root, Address(":a:**")).matches()) == [
        tree.aa,
        tree.ab,
        tree.aba,
        tree.abb,
        tree.abc,
    ]
    assert list(resolve_address(space, tree.root, Address(":a:b:**")).matches()) == [tree.aba, tree.abb, tree.abc]
    assert list(resolve_address(space, tree.root, Address(":**:a")).matches()) == [tree.a, tree.aa, tree.aba]


def test__resolve_address__does_not_fail_on_optional_element() -> None:
    tree = ExampleNodeTree()
    space = NodeAddressSpace(tree.root)

    # We can sucessfully resolve :a:b? to :a:b.
    assert list(resolve_address(space, tree.root, Address(":a:b?")).matches()) == [tree.ab]

    # We can sucessfully resolve :a:dontexist? to nothing, as the address does not exist.
    assert list(resolve_address(space, tree.root, Address(":a:dontexist?")).matches()) == []

    # Try to resolve :a:dontexist without an optional element fails.
    with raises(AddressResolutionError):
        list(resolve_address(space, tree.root, Address(":a:dontexist")).matches())

    # We can sucessfully resolve :a:b?:a to :a:b:a.
    assert list(resolve_address(space, tree.root, Address(":a:b?:a")).matches()) == [tree.aba]

    # We can sucessfully resolve :a:b?:dontexist to nothing, as the element that does not exist is optional.
    assert list(resolve_address(space, tree.root, Address(":a:b:dontexist?")).matches()) == []

    # It is also okay to resolve :a::missing1?:missing to nothing, because we won't try to look up the missing2
    # on missing1, since missing1 doesn't exist in the first place.
    assert list(resolve_address(space, tree.root, Address(":a:missing1?:missing2")).matches()) == []

    # However, trying to lookup a non-optional missing element on an optional but existing element fails.
    with raises(AddressResolutionError):
        list(resolve_address(space, tree.root, Address(":a:b?:dontexist")).matches())



import termcolor

from kraken.core.cli.executor import COLORS_BY_STATUS
from kraken.core.system.task import TaskStatusType


def test__COLORS_BY_STATUS__contains_a_color_for_every_status() -> None:
    for key in TaskStatusType:
        assert key in COLORS_BY_STATUS
        assert COLORS_BY_STATUS[key] in termcolor.COLORS
